/* Tests dynamically extending MzScheme's syntax */

/* A "window" is a list of at least 4 items */

#include "escheme.h"

Scheme_Object *with_win_execute(Scheme_Object *cdata, Scheme_Env *env);

Scheme_Object *window_get_name(Scheme_Object *w)
{
  if (SCHEME_PAIRP(w))
    return SCHEME_CAR(w);
  else
    return scheme_false;
}

Scheme_Object *window_get_width(Scheme_Object *w)
{
  if (SCHEME_PAIRP(w) && SCHEME_PAIRP(SCHEME_CDR(w)))
    return SCHEME_CADR(w);
  else
    return scheme_false;
}

Scheme_Object *window_get_height(Scheme_Object *w)
{
  if (SCHEME_PAIRP(w) && SCHEME_PAIRP(SCHEME_CDR(w))
      && SCHEME_PAIRP(SCHEME_CDR(SCHEME_CDR(w))))
    return SCHEME_CADR(SCHEME_CDR(w));
  else
    return scheme_false;
}

Scheme_Object *window_get_dc(Scheme_Object *w)
{
  if (SCHEME_PAIRP(w) && SCHEME_PAIRP(SCHEME_CDR(w))
      && SCHEME_PAIRP(SCHEME_CDR(SCHEME_CDR(w)))
      && SCHEME_PAIRP(SCHEME_CDR(SCHEME_CDR(SCHEME_CDR(w)))))
    return SCHEME_CADR(SCHEME_CDR(SCHEME_CDR(w)));
  else
    return scheme_false;
}

/* Constant values: */
Scheme_Type with_win_data_type;
Scheme_Object *name_symbol, *width_symbol, *height_symbol, *dc_symbol;

/* Compiled info structure: */
typedef struct {
  Scheme_Type type;
  int allocate_frame;
  int start_frame_slot;
  int extra_frame_slots;
  Scheme_Object *win_expr, *body;
} With_Win_Data;

Scheme_Object *with_win_syntax(Scheme_Object *form, Scheme_Env *env, 
                               Scheme_Compile_Info *rec)
{
  Scheme_Object *body, *win_expr;
  With_Win_Data *data;
  Scheme_Env *new_env;
  Scheme_Compile_Info subrecs[2];

  /* We really need form-checking here: */
  form = SCHEME_CDR(form);
  if (!SCHEME_PAIRP(form))
    scheme_signal_error("with-window: bad form");
  win_expr = SCHEME_CAR(form);
  body = SCHEME_CDR(form);

  data = (With_Win_Data *)scheme_malloc(sizeof(With_Win_Data));
  data->type = with_win_data_type;

  /* Initialize sub-compilation info; we have 2 sub-parts: */
  scheme_init_compile_recs(rec, subrecs, 2);

  /* Compile the expression */
  data->win_expr = scheme_compile_expr(win_expr, env, &subrecs[0]);

  if (rec->base_let_depth >= 0) {
    /* Someone else will allocate the environment frame at run-time: */
    data->allocate_frame = 0;
    data->start_frame_slot = rec->base_let_depth;

    /* Re-allocate last frame, making room for 4 more: */
    new_env = scheme_lengthen_compilation_frame(4, env);
  } else {
    /* We must allocate the environment frame: */
    data->allocate_frame = 1;
    data->start_frame_slot = 0;
    
    new_env = scheme_new_compilation_frame(4, 0);
    new_env = scheme_extend_env(new_env, env);
  }
  
  /* Add names for our locals: */
  scheme_add_compilation_binding(data->start_frame_slot, name_symbol, 
				 new_env);
  scheme_add_compilation_binding(data->start_frame_slot + 1, width_symbol, 
				 new_env);
  scheme_add_compilation_binding(data->start_frame_slot + 2, height_symbol, 
				 new_env);
  scheme_add_compilation_binding(data->start_frame_slot + 3, dc_symbol, 
				 new_env);

  /* Make an adjustment to the second sub-rec: account for 4 locals */
  subrecs[1].base_let_depth = data->start_frame_slot + 4;

  /* Compile the body: */
  data->body = scheme_compile_sequence(body, new_env, &subrecs[1]);

  /* Merge compilation information: */
  scheme_merge_compile_recs(rec, subrecs, 2);

  /* If we allocate, did sub-expressions ask for extra frame slots?: */
  if (data->allocate_frame)
    data->extra_frame_slots = rec->max_let_depth;

  if (!data->allocate_frame)
    /* We made a copy of env; propogate the settable flag: */
    scheme_frame_backinfo(new_env, env);

  /* Adjust outgoing info to account for 4 locals: */
  rec->max_let_depth += 4;

  return scheme_make_syntax_compilation(with_win_execute, 
                                        (Scheme_Object *)data);
}

Scheme_Object *with_win_execute(Scheme_Object *cdata, Scheme_Env *env)
{
  With_Win_Data *data;
  Scheme_Object *win;
 
  data = (With_Win_Data *)cdata;

  /* First, evaluate the initial expression to get a window: */
  win = scheme_eval_compiled(data->win_expr, env);

  /* Allocate a frame, if it was not done for us: */
  if (data->allocate_frame) {
    Scheme_Env *frame;

    /* Allocate a new the frame: */
    frame = scheme_new_frame(4 + data->extra_frame_slots);
    env = scheme_extend_env(frame, env);
  }

  /* Assume all the functions window_get_name, etc. are defined,
     taking a Scheme window and returning a scheme value */
  scheme_add_binding(data->start_frame_slot,
                     window_get_name(win), env);
  scheme_add_binding(data->start_frame_slot + 1,
                     window_get_height(win), env);
  scheme_add_binding(data->start_frame_slot + 2,
                     window_get_width(win), env);
  scheme_add_binding(data->start_frame_slot + 3,
                     window_get_dc(win), env);

  /* Evaluate the body to get the final value: */
  return scheme_tail_eval_expr(data->body, env);
}

Scheme_Object *with_win_expand(Scheme_Object *form, Scheme_Env *env, 
                               int depth)
{
  Scheme_Object *body, *win_expr;

  /* We really need form-checking here: */
  win_expr = SCHEME_CAR(SCHEME_CDR(form));
  body = SCHEME_CDR(SCHEME_CDR(form));

  /* Expand window expression: */
  win_expr = scheme_expand_expr(win_expr, env, depth);
  
  /* Don't worry about putting everything in the same frame for expansion: */
  env = scheme_add_compilation_frame(scheme_make_pair(name_symbol,
                           scheme_make_pair(width_symbol,
                             scheme_make_pair(height_symbol,
                               scheme_make_pair(dc_symbol, scheme_null)))),
                         env, 
                         (depth < 0) ? SCHEME_ANNOTATING : 0);

  /* Expand body: */
  body = scheme_expand_block(body, env, depth);

  return scheme_make_pair(SCHEME_CAR(form),
                          scheme_make_pair(win_expr, body));
}

Scheme_Object *write_with_wind_data(Scheme_Object *cdata)
/* Bundles compiled data: */
{
  With_Win_Data *data;
 
  data = (With_Win_Data *)cdata;

  return scheme_make_pair(data->allocate_frame ? scheme_true : scheme_false,
           scheme_make_pair(scheme_make_integer(data->start_frame_slot),
             scheme_make_pair(scheme_make_integer(data->extra_frame_slots),
               scheme_make_pair(data->win_expr, data->body))));
}

Scheme_Object *read_with_wind_data(Scheme_Object *form)
/* Unbundles compiled data: */
{
  With_Win_Data *data;
 
  data = (With_Win_Data *)scheme_malloc(sizeof(With_Win_Data));

  data->allocate_frame = !SCHEME_FALSEP(SCHEME_CAR(form));
  form = SCHEME_CDR(form);
  data->start_frame_slot = SCHEME_INT_VAL(SCHEME_CAR(form));
  form = SCHEME_CDR(form);
  data->extra_frame_slots = SCHEME_INT_VAL(SCHEME_CAR(form));
  form = SCHEME_CDR(form);
  data->win_expr = SCHEME_CAR(form);
  data->body = SCHEME_CDR(form);

  return (Scheme_Object *)data;
}

void scheme_initialize(Scheme_Env *global_env)
{
  with_win_data_type = scheme_make_type("<with-win-data>");

  name_symbol = scheme_intern_symbol("name");
  width_symbol = scheme_intern_symbol("width");
  height_symbol = scheme_intern_symbol("height");
  dc_symbol = scheme_intern_symbol("drawing-context");

  scheme_register_syntax_executer("withwin", with_win_execute);
  scheme_add_global("with-window", 
                    scheme_make_compiled_syntax(with_win_syntax, 
                                                with_win_expand),
                    global_env);

  scheme_install_type_writer(with_win_data_type, write_with_wind_data);
  scheme_install_type_reader(with_win_data_type, read_with_wind_data);

  scheme_register_extension_global(&name_symbol, sizeof(Scheme_Object *));
  scheme_register_extension_global(&width_symbol, sizeof(Scheme_Object *));
  scheme_register_extension_global(&height_symbol, sizeof(Scheme_Object *));
  scheme_register_extension_global(&dc_symbol, sizeof(Scheme_Object *));
}
