/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"
#include <string.h>

/* global_constants */
Scheme_Object *scheme_true;
Scheme_Object *scheme_false;
Scheme_Object *scheme_not_prim;

/* locals */
static Scheme_Object *scheme_make_true (void);
static Scheme_Object *scheme_make_false (void);
static Scheme_Object *not_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *boolean_p_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *eq_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *eqv_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *equal_prim (int argc, Scheme_Object *argv[]);

static int list_equal (Scheme_Object *lst1, Scheme_Object *lst2);
static int vector_equal (Scheme_Object *vec1, Scheme_Object *vec2);

void
scheme_init_bool (Scheme_Env *env)
{
  if (scheme_starting_up) {
    scheme_true = scheme_make_true ();
    scheme_false = scheme_make_false ();
    scheme_not_prim = scheme_make_prim_w_arity(not_prim, 1, 1);
  }

  scheme_add_global_constant("not", scheme_not_prim, env);
  scheme_add_global_constant("boolean?", 
			     scheme_make_prim_w_arity(boolean_p_prim, 1, 1), 
			     env);
  scheme_add_global_constant("eq?", 
			     scheme_make_prim_w_arity(eq_prim, 2, 2), 
			     env);
  scheme_add_global_constant("eqv?", 
			     scheme_make_prim_w_arity(eqv_prim, 2, 2), 
			     env);
  scheme_add_global_constant("equal?", 
			     scheme_make_prim_w_arity(equal_prim, 2, 2),
			     env);
}

static Scheme_Object *
scheme_make_true (void)
{
  Scheme_Object *truev;

  truev = scheme_alloc_eternal_object();
  truev->type = scheme_true_type;
  return (truev);
}

static Scheme_Object *
scheme_make_false (void)
{
  Scheme_Object *falsev;

  falsev = scheme_alloc_eternal_object();
  falsev->type = scheme_false_type;
  return (falsev);
}

static Scheme_Object *
not_prim (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("not", 1, 1, argc, argv);

  return (SAME_OBJ(argv[0], scheme_false) ? scheme_true : scheme_false);
}

static Scheme_Object *
boolean_p_prim (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("boolean?", 1, 1, argc, argv);

  return ((SAME_OBJ(argv[0], scheme_false) || SAME_OBJ(argv[0], scheme_true))
	? scheme_true : scheme_false);  
}

static Scheme_Object *
eq_prim (int argc, Scheme_Object *argv[])
{
  if (argc != 2)
    scheme_wrong_count("eq?", 2, 2, argc, argv);

  return (SAME_OBJ(argv[0], argv[1]) ? scheme_true : scheme_false);
}

static Scheme_Object *
eqv_prim (int argc, Scheme_Object *argv[])
{
  if (argc != 2)
    scheme_wrong_count("eqv?", 2, 2, argc, argv);

  return ((scheme_eqv (argv[0], argv[1])) ? scheme_true : scheme_false);
}

static Scheme_Object *
equal_prim (int argc, Scheme_Object *argv[])
{
  if (argc != 2)
    scheme_wrong_count("equal?", 2, 2, argc, argv);

  return ((scheme_equal (argv[0], argv[1])) ? scheme_true : scheme_false);
}

int 
scheme_eq (Scheme_Object *obj1, Scheme_Object *obj2)
{
  return SAME_OBJ(obj1, obj2);
}

int
scheme_eqv (Scheme_Object *obj1, Scheme_Object *obj2)
{
  if (SAME_OBJ(obj1, obj2))
    return 1;
  else if (NOT_SAME_TYPE(SCHEME_TYPE(obj1), SCHEME_TYPE(obj2)))
    return 0;
#ifndef FAST_NUMBERS
  else if (SCHEME_INTP(obj1))
    return SCHEME_INT_VAL(obj1) == SCHEME_INT_VAL(obj2);
#endif
  else if (SCHEME_DBLP(obj1))
    return SCHEME_DBL_VAL(obj1) == SCHEME_DBL_VAL(obj2);
  else if (SCHEME_BIGNUMP(obj1))
    return scheme_bignum_eq(obj1, obj2);
  else if (SCHEME_CHARP(obj1))
    return SCHEME_CHAR_VAL(obj1) == SCHEME_CHAR_VAL(obj2);
  else
    return 0;
}

int
scheme_equal (Scheme_Object *obj1, Scheme_Object *obj2)
{
  if (scheme_eqv (obj1, obj2))
    return 1;
  else if (NOT_SAME_TYPE(SCHEME_TYPE(obj1), SCHEME_TYPE(obj2)))
      return 0;
  else if (SCHEME_PAIRP(obj1))
    return list_equal(obj1, obj2);
  else if (SCHEME_VECTORP(obj1))
    return vector_equal(obj1, obj2);
  else if (SCHEME_STRINGP(obj1))
    return !strcmp(SCHEME_STR_VAL(obj1), SCHEME_STR_VAL(obj2));
  else  if (SAME_TYPE(SCHEME_TYPE(obj1), scheme_structure_type))
    return scheme_equal_structs(obj1, obj2);
  else
    return 0;
}

static int
list_equal (Scheme_Object *lst1, Scheme_Object *lst2)
{
  if (SCHEME_NULLP(lst1) && SCHEME_NULLP(lst2))
    return 1;
  else
    {
      return (scheme_equal(SCHEME_CAR (lst1), SCHEME_CAR (lst2)) &&
	      scheme_equal(SCHEME_CDR (lst1), SCHEME_CDR (lst2)));
    }
}

static int
vector_equal (Scheme_Object *vec1, Scheme_Object *vec2)
{
  int i;

  for ( i=0 ; i<SCHEME_VEC_SIZE(vec1) ; ++i )
    {
      if (! scheme_equal (SCHEME_VEC_ELS(vec1)[i], SCHEME_VEC_ELS(vec2)[i]))
	{
	  return 0;
	}
    }
  return 1;
}



