/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"
#ifdef TIME_SYNTAX
#ifdef USE_FTIME
#include <sys/timeb.h>
#else
#include <sys/time.h>
#endif
#endif

/* globals */
int scheme_escape_continuations_only; 

Scheme_Object *scheme_void;
Scheme_Object *scheme_void_func;

Scheme_Object *scheme_tail_call_waiting;
Scheme_Object *scheme_tail_rator;
int scheme_tail_num_rands;
Scheme_Object **scheme_tail_rands;
int scheme_multiple_count;
Scheme_Object **scheme_multiple_array;

#ifndef ERROR_ON_OVERFLOW
Scheme_Object *scheme_overflow_reply;
Scheme_Jumpup_Buf scheme_overflow_cont;
#endif

/* locals */
static Scheme_Object *procedure_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *apply (int argc, Scheme_Object *argv[]);
static Scheme_Object *map (int argc, Scheme_Object *argv[]);
static Scheme_Object *for_each (int argc, Scheme_Object *argv[]);
static Scheme_Object *call_ec (int argc, Scheme_Object *argv[]);
static Scheme_Object *call_cc (int argc, Scheme_Object *argv[]);
static Scheme_Object *void_func (int argc, Scheme_Object *argv[]);
static Scheme_Object *is_void_func (int argc, Scheme_Object *argv[]);
static Scheme_Object *dynamic_wind (int argc, Scheme_Object *argv[]);
static Scheme_Object *rep (int argc, Scheme_Object *argv[]);
#ifdef TIME_SYNTAX
static Scheme_Object *time_apply(int argc, Scheme_Object *argv[]);
#endif
static Scheme_Object *arity (int argc, Scheme_Object *argv[]);
static Scheme_Object *call_with_values(int argc, Scheme_Object *argv[]);
Scheme_Object *scheme_values(int argc, Scheme_Object *argv[]);

static Scheme_Object *write_compiled_closure(Scheme_Object *obj);
static Scheme_Object *read_compiled_closure(Scheme_Object *obj);

static Scheme_Object *rep_prim;

#define CONS(a,b) scheme_make_pair(a,b)

#define DYNAMIC_WIND "dynamic-wind"

void
scheme_init_fun (Scheme_Env *env)
{
  Scheme_Object *o;

  if (scheme_starting_up) {
    scheme_void = scheme_alloc_eternal_object();
    scheme_void->type = scheme_void_type;

    scheme_tail_call_waiting = scheme_alloc_eternal_object();
    scheme_tail_call_waiting->type = scheme_tail_call_waiting_type;

    scheme_void_func = scheme_make_prim_w_arity(void_func, 0, -1);
  }

  scheme_add_global_constant("procedure?", 
			     scheme_make_prim_w_arity(procedure_p, 1, 1), 
			     env);
  scheme_add_global_constant("apply", 
			     scheme_make_prim_w_arity(apply, 2, -1), 
			     env);
  scheme_add_global_constant("map", 
			     scheme_make_prim_w_arity(map, 2, -1), 
			     env);
  scheme_add_global_constant("for-each", 
			     scheme_make_prim_w_arity(for_each, 2, -1), 
			     env);
  scheme_add_global_constant("call-with-values",
			     scheme_make_prim_w_arity(call_with_values, 2, 2),
			     env);
  scheme_add_global_constant("values",
			     scheme_make_prim_w_arity(scheme_values, 0, -1),
			     env);
  scheme_add_global_constant("call-with-escaping-continuation", 
			     o = scheme_make_prim_w_arity(call_ec, 1, 1), 
			     env);
  scheme_add_global_constant("call/ec", o, env);  
  scheme_add_global_constant("call-with-current-continuation", 
			     o = scheme_make_prim_w_arity(scheme_escape_continuations_only ? call_ec : call_cc, 1, 1), 
			     env);
  scheme_add_global_constant("call/cc", o, env);  
  scheme_add_global_constant("void", scheme_void_func, env);  
  scheme_add_global_constant("void?", 
			     scheme_make_prim_w_arity(is_void_func, 1, 1), 
			     env);  
#ifdef TIME_SYNTAX
  scheme_add_global_constant("time-apply", 
			     scheme_make_prim_w_arity(time_apply, 1, 1), 
			     env);
#endif

  scheme_add_global_constant("read-eval-print-loop", 
			      rep_prim = scheme_make_prim_w_arity(rep, 0, 0), 
			     env);  

  scheme_add_global_constant(DYNAMIC_WIND, 
			     scheme_make_prim_w_arity(dynamic_wind, 3, 3), 
			     env);

  /* scheme_add_global_constant("void-value", scheme_void, env); */

  scheme_add_global_constant("arity", 
			     scheme_make_prim_w_arity(arity, 1, 1), 
			     env);

  if (scheme_starting_up) {
    scheme_install_type_writer(scheme_closure_compilation_type,
			       write_compiled_closure);
    scheme_install_type_reader(scheme_closure_compilation_type,
			       read_compiled_closure);
  }
}

Scheme_Object *
scheme_value(Scheme_Object *obj)
{
  if (SAME_OBJ(obj, scheme_tail_call_waiting)) {
    return _scheme_apply(scheme_tail_rator, scheme_tail_num_rands, scheme_tail_rands);
  } else if (SAME_OBJ(obj, scheme_eval_waiting))
    return _scheme_eval_compiled(scheme_eval_wait_expr, scheme_eval_wait_env);
  else if (obj)
    return obj;
  else
    return scheme_void;
}

Scheme_Object *
scheme_make_prim_w_arity_and_eternal(Scheme_Prim *fun, int eternal,
				     short mina, short maxa)
{
  Scheme_Primitive_Proc *prim;

  if (eternal)
    prim = (Scheme_Primitive_Proc *)scheme_malloc_eternal(sizeof(Scheme_Primitive_Proc));
  else
    prim = (Scheme_Primitive_Proc *)scheme_malloc_stubborn(sizeof(Scheme_Primitive_Proc));
  prim->type = scheme_prim_type;
  SCHEME_PRIM(prim) = fun;
  prim->mina = mina;
  prim->maxa = maxa;

  if (!eternal)
    scheme_end_stubborn_change(prim);

  return (Scheme_Object *)prim;
}

Scheme_Object *scheme_make_prim(Scheme_Prim *fun)
{
  return scheme_make_prim_w_arity_and_eternal(fun, 1, 0, -1);
}

Scheme_Object *
scheme_make_noneternal_prim (Scheme_Prim *fun)
{
  return scheme_make_prim_w_arity_and_eternal(fun, 0, 0, -1);
}

Scheme_Object *
scheme_make_prim_w_arity(Scheme_Prim *fun, short mina, short maxa)
{
  return scheme_make_prim_w_arity_and_eternal(fun, 1, mina, maxa);
}

Scheme_Object *
scheme_make_noneternal_prim_w_arity(Scheme_Prim *fun, short mina, short maxa)
{
  return scheme_make_prim_w_arity_and_eternal(fun, 0, mina, maxa);
}

Scheme_Object *
scheme_make_closed_prim_w_arity(Scheme_Closed_Prim *fun, void *data,
				short mina, short maxa)
{
  Scheme_Closed_Primitive_Proc *prim;

  prim = (Scheme_Closed_Primitive_Proc *)scheme_malloc(sizeof(Scheme_Closed_Primitive_Proc));

  prim->type = scheme_closed_prim_type;
  SCHEME_CLSD_PRIM(prim) = fun;
  SCHEME_CLSD_PRIM_DATA(prim) = data;
  prim->mina = mina;
  prim->maxa = maxa;
  
  return (Scheme_Object *)prim;
}

Scheme_Object *
scheme_make_closed_prim(Scheme_Closed_Prim *fun, void *data)
{
  return scheme_make_closed_prim_w_arity(fun, data, 0, -1);
}

Scheme_Object *
scheme_make_closure (Scheme_Env *env, Scheme_Object *code)
{
  Scheme_Object *closure;

  closure = scheme_alloc_stubborn_object ();

  closure->type = scheme_closure_type;
  SCHEME_CLOS_ENV (closure) = env;
  SCHEME_CLOS_CODE (closure) = code;

  scheme_end_stubborn_change((void *)closure);

  return closure;
}

#define CL_STUBBORN_ON 0

Scheme_Object *
scheme_make_compiled_closure (Scheme_Env *env, Scheme_Object *compiled_code)
{
  Scheme_Object *closure;

#if CL_STUBBORN_ON
  closure = scheme_alloc_stubborn_object ();
#else
  closure = scheme_alloc_object();
#endif

  closure->type = scheme_compiled_closure_type;
  SCHEME_CLOS_ENV (closure) = env;
  SCHEME_CLOS_CODE (closure) = compiled_code;

#if CL_STUBBORN_ON
  scheme_end_stubborn_change((void *)closure);
#endif

  return closure;
}

Scheme_Object *
scheme_make_closure_compilation (Scheme_Env *env, Scheme_Object *code, 
				 Scheme_Compile_Info *rec)
{
  Scheme_Object *allparams, *params, *forms, *param, *name, *rest;
  Scheme_Closure_Compilation_Data *data;
  Scheme_Env *frame;
  int i;

  data  = (Scheme_Closure_Compilation_Data *)scheme_malloc_stubborn(sizeof(Scheme_Closure_Compilation_Data));

  data->type = scheme_closure_compilation_type;

  /* data->src = code; */

  params = allparams = /* data->params = */ SCHEME_CAR(code);

  data->num_params = 0;
  for (; SCHEME_PAIRP(params); params = SCHEME_CDR(params))
    data->num_params++;
  data->has_rest = !SCHEME_NULLP(params);
  if (data->has_rest)
    data->num_params++;

  forms = SCHEME_CDR(code);

  /* Add frame if # params > 0 */
  /* !env->next is a hack: adds a useless frame to block illegal
     (define ...) statements */
  if (data->num_params || !env->basic.next) {

    /* Check for duplicate names: */
    for (params = allparams;
	 SCHEME_PAIRP(params); 
	 params = SCHEME_CDR(params)) {
      name = SCHEME_CAR(params);
      for (rest = SCHEME_CDR(params); 
	   !SCHEME_NULLP(rest); 
	   rest = SCHEME_CDR(rest)) {
	if (!SCHEME_PAIRP(rest))
	  param = rest;
	else
	  param = SCHEME_CAR(rest);
	if (SCHEME_SYMBOLP(name) && SAME_OBJ(param, name)) {
	  scheme_signal_error("duplicate argument name \"%s\" "
			      "in `lambda' form",
			      SCHEME_SYM_VAL(name));
	}
	if (!SCHEME_PAIRP(rest))
	  break;
      }
    }

    frame = scheme_new_compilation_frame(data->num_params, 0);
    params = allparams;
    for (i = 0; i < data->num_params; i++) {
      if (!SCHEME_PAIRP(params))
	param = params;
      else
	param = SCHEME_CAR(params);
      scheme_check_identifier("lambda", param, NULL, env);
      scheme_add_binding (i, param, frame);
      if (SCHEME_PAIRP(params))
	params = SCHEME_CDR (params);
    }
    frame = scheme_extend_env (frame, env);
  } else
    frame = env;

  if (SCHEME_NULLP(forms))
    scheme_signal_error("syntax error: no forms in body");

  if (!data->num_params)
    rec->base_let_depth = -1; /* enclosed lets must allocate a frame */
  else
    rec->base_let_depth = data->num_params;

  data->code = scheme_compile_sequence(forms, frame, rec);

  if (!data->num_params)
    data->extra_frame_slots = 0;
  else
    data->extra_frame_slots = rec->max_let_depth;
    
  data->must_allocate = (rec->can_return_closure 
			 || (frame->basic.flags & SCHEME_HAS_SET_BANG));

  rec->max_let_depth = 0;
  rec->can_return_closure = 1;

  scheme_end_stubborn_change((void *)data);

  if (!env->basic.next && rec->can_optimize_constants)
    /* If it's just at global level, go ahead and finialize closure */
    return scheme_make_compiled_closure(env, (Scheme_Object *)data);
  else
    return (Scheme_Object *)data;
}

Scheme_Object *scheme_make_econt ()
{
  Scheme_Object *cont;

  cont = scheme_malloc(sizeof(Scheme_Escaping_Cont));
  cont->type = scheme_escaping_cont_type;
  SCHEME_CONT_VAL(cont) = NULL;
  SCHEME_CONT_HOME(cont) = NULL;
  
  return cont;
}

Scheme_Object *scheme_make_ccont ()
{
  Scheme_Cont *cont;

  cont = (Scheme_Cont *)scheme_malloc(sizeof(Scheme_Cont));
  cont->type = scheme_cont_type;
  scheme_init_jmpup_buf(&cont->buf);
  cont->ok = scheme_current_process->cc_ok;
  cont->dw = scheme_current_process->dw;
  cont->home = scheme_current_process;
  cont->orig_overflow = scheme_current_process->stack_overflow;
  cont->save_overflow = scheme_current_process->overflow;
  scheme_save_env_stack(cont->envss);

  return (Scheme_Object *)cont;
}

Scheme_Object *
scheme_top_level_eval(Scheme_Object *rator,int num_rands,Scheme_Object **rands)
{
  Scheme_Object *v;
  long *old_cc_ok, *ok;
  void *old_cc_start;
  jmp_buf save, oversave;
  int orig_overflow;
  Scheme_Env_Stack_State envss;

  if (scheme_active_but_sleeping)
    scheme_wake_up();

  old_cc_ok = scheme_current_process->cc_ok;
  old_cc_start = scheme_current_process->cc_start;

  scheme_current_process->cc_ok = 
    ok = (long *)scheme_malloc_atomic(sizeof(long));
  scheme_current_process->cc_start = &v;
  
  if (old_cc_ok)
    *old_cc_ok = 0;
  *ok = 1;

  orig_overflow = scheme_current_process->stack_overflow;
  scheme_save_env_stack(envss);

#ifndef ERROR_ON_OVERFLOW
  memcpy(&oversave, &scheme_current_process->overflow_buf, sizeof(jmp_buf));
  if (scheme_setjmp(scheme_current_process->overflow_buf)) {
    Scheme_Overflow *overflow;

    overflow = (Scheme_Overflow *)scheme_malloc(sizeof(Scheme_Overflow));

    memcpy(&overflow->cont, &scheme_overflow_cont, 
	   sizeof(Scheme_Jumpup_Buf));
    overflow->prev = scheme_current_process->overflow;
    scheme_current_process->overflow = overflow;

    memcpy(&overflow->savebuf, &scheme_error_buf, sizeof(jmp_buf));
    if (scheme_setjmp(scheme_error_buf)) {
      scheme_overflow_reply = NULL; /* means "continue the error" */
    } else {
      scheme_overflow_reply = 
	_scheme_apply(scheme_tail_rator, scheme_tail_num_rands, 
		      scheme_tail_rands);
    }

    overflow = scheme_current_process->overflow;
    memcpy(&scheme_error_buf, &overflow->savebuf, sizeof(jmp_buf));
    scheme_current_process->overflow = overflow->prev;
    memcpy(&scheme_overflow_cont, &overflow->cont, 
	   sizeof(Scheme_Jumpup_Buf));
    overflow = NULL; /* Maybe helps GC */
    scheme_longjmpup(&scheme_overflow_cont);
  }
#endif

  memcpy(&save, &scheme_error_buf, sizeof(jmp_buf));
  if (scheme_setjmp(scheme_error_buf)) {
    scheme_restore_env_stack(envss);
    *ok = 0;
    if (old_cc_ok)
      *old_cc_ok = 1;
    scheme_current_process->cc_ok = old_cc_ok;
    scheme_current_process->cc_start = old_cc_start;  
    scheme_current_process->stack_overflow = orig_overflow;
    scheme_longjmp(save, 1);
  }

  if (num_rands >= 0)
    v = _scheme_apply(rator, num_rands, rands);
  else
    v = _scheme_eval_compiled(rator, (Scheme_Env *)rands);

  memcpy(&scheme_error_buf, &save, sizeof(jmp_buf));
#ifndef ERROR_ON_OVERFLOW
  memcpy(&scheme_current_process->overflow_buf, &oversave, sizeof(jmp_buf));
#endif

  scheme_current_process->stack_overflow = orig_overflow;

  *ok = 0;
  if (old_cc_ok)
    *old_cc_ok = 1;
  scheme_current_process->cc_ok = old_cc_ok;
  scheme_current_process->cc_start = old_cc_start;

  if (scheme_active_but_sleeping)
    scheme_wake_up();

  return v;
}

Scheme_Object *
scheme_apply(Scheme_Object *rator, int num_rands, Scheme_Object **rands)
{
  return scheme_top_level_eval(rator, num_rands, rands);
}

Scheme_Object *
scheme_tail_apply (Scheme_Object *rator, int num_rands, Scheme_Object **rands)
{
  int i;

  scheme_tail_rator = rator;
  scheme_tail_num_rands = num_rands;
  
  scheme_tail_rands = scheme_malloc(num_rands * sizeof(Scheme_Object *));
  for (i = num_rands; i--; )
    scheme_tail_rands[i] = rands[i];

  return scheme_tail_call_waiting;
}

static
Scheme_Object *
X_scheme_apply_to_list (Scheme_Object *rator, Scheme_Object *rands, int force,
			int top_level)
{
  int num_rands, i;
  Scheme_Object **rands_vec;

  num_rands = scheme_list_length (rands);
  rands_vec = (Scheme_Object **)scheme_malloc(num_rands 
					      * sizeof(Scheme_Object *));
  for ( i=0 ; i<num_rands ; ++i )
    {
      SCHEME_ASSERT(SCHEME_PAIRP(rands), "bad application form");
      rands_vec[i] = SCHEME_CAR (rands);
      rands = SCHEME_CDR (rands);
    }
  
  if (top_level)  {
    if (force)
      return scheme_apply(rator, num_rands, rands_vec);
    else
      return scheme_tail_apply(rator, num_rands, rands_vec);
  } else {
    if (force)
      return _scheme_apply(rator, num_rands, rands_vec);
    else
      return _scheme_tail_apply(rator, num_rands, rands_vec);
  }
}

Scheme_Object *
scheme_apply_to_list (Scheme_Object *rator, Scheme_Object *rands)
{
  return X_scheme_apply_to_list(rator, rands, 1, 1);
}

Scheme_Object *
scheme_tail_apply_to_list (Scheme_Object *rator, Scheme_Object *rands)
{
  return X_scheme_apply_to_list(rator, rands, 0, 1);
}

Scheme_Object *
_scheme_apply_to_list (Scheme_Object *rator, Scheme_Object *rands)
{
  return X_scheme_apply_to_list(rator, rands, 1, 0);
}

Scheme_Object *
_scheme_tail_apply_to_list (Scheme_Object *rator, Scheme_Object *rands)
{
  return X_scheme_apply_to_list(rator, rands, 0, 0);
}

/* locals */

static Scheme_Object *
procedure_p (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("procedure?", 1, 1, argc, argv);
  return (SCHEME_PROCP(argv[0]) ? scheme_true : scheme_false);
}

static Scheme_Object *make_arity(short mina, short maxa)
{
  if (mina == maxa)
    return scheme_make_integer(mina);
  else if (maxa == -1) {
    Scheme_Object *p[1];
    p[0] = scheme_make_integer(mina);
    return _scheme_apply(scheme_arity_at_least_constr, 1, p);
  } else {
    int i;
    Scheme_Object *l = scheme_null;

    for (i = maxa; i >= mina; --i)
      l = scheme_make_pair(scheme_make_integer(i), l);

    return l;
  }
}

static Scheme_Object *arity (int argc, Scheme_Object *argv[])
{
  Scheme_Type type;
  short mina, maxa;

  if (argc != 1)
    scheme_wrong_count("arity", 1, 1, argc, argv);
  if (!SCHEME_PROCP(argv[0]))
    scheme_wrong_type("arity", "procedure", 0, argc, argv);

  type = SCHEME_TYPE(argv[0]);
  if (type == scheme_prim_type) {
    mina = ((Scheme_Primitive_Proc *)argv[0])->mina;
    maxa = ((Scheme_Primitive_Proc *)argv[0])->maxa;
  } else if (type == scheme_closed_prim_type) {
    mina = ((Scheme_Closed_Primitive_Proc *)argv[0])->mina;
    maxa = ((Scheme_Closed_Primitive_Proc *)argv[0])->maxa;
  } else if (type == scheme_cont_type || type == scheme_escaping_cont_type)
    mina = maxa = 1;
  else if (type == scheme_struct_proc_type)
    scheme_struct_get_arity(argv[0], &mina, &maxa);
  else if (type == scheme_case_closure_type) {
    Scheme_Sequence *seq;
    Scheme_Closure_Compilation_Data *data;
    int i;
    Scheme_Object *first = NULL, *last = NULL, *v;

    seq = (Scheme_Sequence *)argv[0];
    for (i = 0; i < seq->count; i++) {
      data = 
	(Scheme_Closure_Compilation_Data *)SCHEME_CLOS_CODE(seq->array[i]);
      mina = maxa = data->num_params;
      if (data->has_rest) {
	--mina;
	maxa = -1;
      }
      
      v = scheme_make_pair(make_arity(mina, maxa), scheme_null);
      if (!last)
	first = v;
      else
	SCHEME_CDR(last) = v;
      last = v;
    }

    if (!first)
      return scheme_null;

    return first;
  } else {
    Scheme_Closure_Compilation_Data *data;

    data = (Scheme_Closure_Compilation_Data *)SCHEME_CLOS_CODE(argv[0]);
    mina = maxa = data->num_params;
    if (data->has_rest) {
      --mina;
      maxa = -1;
    }
  }

  return make_arity(mina, maxa);
}

static Scheme_Object *
apply (int argc, Scheme_Object *argv[])
{
  Scheme_Object *rands, *r;
  Scheme_Object **rand_vec;
  int i, num_rands;

  if (argc < 2)
    scheme_wrong_count("apply", 2, -1, argc, argv);

  if (!SCHEME_PROCP(argv[0]))
    scheme_wrong_type("apply", "procedure", 0, argc, argv);

  rands = argv[argc-1];

  num_rands = (argc - 2);
  r = rands;
  while (!SCHEME_NULLP(r)) {
    if (!SCHEME_PAIRP(r))
      scheme_wrong_type("apply", "proper list", argc - 1, argc, argv);
    r = SCHEME_CDR(r);
    num_rands++;
  }
  rand_vec = (Scheme_Object **)scheme_malloc(num_rands *
					     sizeof(Scheme_Object *));

  for (i = argc - 2; i--; )
    rand_vec[i] = argv[i + 1];
  
  for (i = argc - 2; SCHEME_PAIRP(rands); i++, rands = SCHEME_CDR(rands))
    rand_vec[i] = SCHEME_CAR(rands);

  scheme_tail_rator = argv[0];
  scheme_tail_rands = rand_vec;
  scheme_tail_num_rands = num_rands;

  return scheme_tail_call_waiting;
}

static Scheme_Object *
do_map (int argc, Scheme_Object *argv[], char *name, int make_result)
{
  int i, size = 0;
  Scheme_Object *pair, *v;
  Scheme_Object *retfirst, *retlast = NULL;
  Scheme_Object **args;

  if (argc < 2)
    scheme_wrong_count(name, 2, -1, argc, argv);
  if (!SCHEME_PROCP(argv[0]))
    scheme_wrong_type(name, "procedure", 0, argc, argv);

  for ( i=1 ; i<argc ; ++i ) {
    if (!SCHEME_LISTP (argv[i]))
      scheme_wrong_type(name, "list", i, argc, argv);

    if (i == 1)
      size = scheme_list_length (argv[i]);
    else if (size != scheme_list_length (argv[i]))
      scheme_signal_error ("%s: all lists must have same size", name);
  }
  
  args = (Scheme_Object **)scheme_malloc((argc - 1) * sizeof(Scheme_Object *));

  if (make_result)
    retfirst = retlast = scheme_null;
  else
    retfirst = scheme_void;

  while (!SCHEME_NULLP(argv[1]))
    {
      /* collect args to apply */
      for ( i=1 ; i<argc ; ++i ) {
	if (!SCHEME_PAIRP(argv[i]))
	  scheme_signal_error("%s: arg %d not a proper list", name, i);
	args[i-1] = SCHEME_CAR(argv[i]);
	argv[i] = SCHEME_CDR (argv[i]);
      }
      v = _scheme_apply(argv[0], argc-1, args);

      if (make_result) {
	pair = scheme_make_pair(v, scheme_null);
	if (SCHEME_NULLP (retfirst))
	  retfirst = retlast = pair;
	else {
	  SCHEME_CDR (retlast) = pair;
	  retlast = pair;
	}
      }
    }

  return (retfirst);
}

static Scheme_Object *
map (int argc, Scheme_Object *argv[])
{
  return do_map(argc, argv, "map", 1);
}

static Scheme_Object *
for_each (int argc, Scheme_Object *argv[])
{
  return do_map(argc, argv, "for-each", 0);
}

static Scheme_Object *call_with_values(int argc, Scheme_Object *argv[])
{
  Scheme_Object *v;

  if (argc != 2)
    scheme_wrong_count("call-with-values", 2, 2, argc, argv);
  if (!SCHEME_PROCP(argv[0]))
    scheme_wrong_type("call-with-values", "procedure", 0, argc, argv);
  if (!SCHEME_PROCP(argv[1]))
    scheme_wrong_type("call-with-values", "procedure", 1, argc, argv);

  v = _scheme_apply_multi(argv[0], 0, NULL);
  if (SAME_OBJ(v, scheme_multiple_values)) {
    scheme_tail_num_rands = scheme_multiple_count;
    scheme_tail_rands = scheme_multiple_array;
    scheme_multiple_array = NULL;
  } else {
    scheme_tail_num_rands = 1;
    scheme_tail_rands=(Scheme_Object **)scheme_malloc(sizeof(Scheme_Object *));
    scheme_tail_rands[0] = v;
  }
  
  scheme_tail_rator = argv[1];

  return scheme_tail_call_waiting;
}

Scheme_Object *scheme_values(int argc, Scheme_Object *argv[])
{
  int i;

  if (argc == 1)
    return argv[0];
  scheme_multiple_count = argc;
  if (argc)
    scheme_multiple_array = 
      (Scheme_Object **)scheme_malloc(argc * sizeof(Scheme_Object *));
  
  for (i = 0; i < argc; i++)
    scheme_multiple_array[i] = argv[i];

  return scheme_multiple_values;
}

static void pre_call_ec(void *ec)
{
  SCHEME_CONT_HOME(ec) = scheme_current_process;
}

static void post_call_ec(void *ec)
{
  SCHEME_CONT_HOME(ec) = NULL;
}

static Scheme_Object *do_call_ec(void *ec)
{
  Scheme_Object *p[1];

  p[0] = (Scheme_Object *)ec;

  return _scheme_apply(((Scheme_Escaping_Cont *)ec)->f, 1, p);
}

static Scheme_Object *handle_call_ec(void *ec)
{
  if (SCHEME_CONT_VAL(ec)) {
    scheme_jumping_to_continuation = 0;
    scheme_current_process->stack_overflow = 
      ((Scheme_Escaping_Cont *)ec)->orig_overflow;
    return SCHEME_CONT_VAL(ec);
  } else
    return NULL;
}

static Scheme_Object *
call_ec (int argc, Scheme_Object *argv[])
{
  Scheme_Escaping_Cont *cont;

  if (argc != 1)
    scheme_wrong_count("call-with-escaping-continuation", 1, 1, argc, argv);
  if (!SCHEME_PROCP(argv[0]))
    scheme_wrong_type("call-with-escaping-continuation", "procedure", 
		      0, argc, argv);

  cont = (Scheme_Escaping_Cont *)scheme_make_econt();
  cont->f = argv[0];
  cont->orig_overflow = scheme_current_process->stack_overflow;

  return scheme_dynamic_wind(pre_call_ec, do_call_ec, post_call_ec,
			     handle_call_ec, (void *)cont);
}

static Scheme_Object *
call_cc (int argc, Scheme_Object *argv[])
{
  Scheme_Object *ret, *p[1];
  Scheme_Cont *cont;
  Scheme_Dynamic_Wind *dw, *start;
  Scheme_Env_Stack *env_stack;
  long total, pos;

  if (argc != 1)
    scheme_wrong_count("call-with-current-continuation", 1, 1, argc, argv);
  if (!SCHEME_PROCP(argv[0]))
    scheme_wrong_type("call-with-current-continuation", "procedure", 
		      0, argc, argv);

  cont = (Scheme_Cont *)scheme_make_ccont();

#if MAX_QUICK_VALUES > 0
  env_stack = scheme_current_process->env_stack;
  total = NUM_QUICK_ENVS_PER_SEGMENT - env_stack->stack_pos;
  while (env_stack->prev) {
    total += NUM_QUICK_ENVS_PER_SEGMENT;
    env_stack = env_stack->prev;
  }

  cont->saved_stack_size = total;
  cont->saved_env_stack = 
    (Scheme_Quick_Env *)scheme_malloc(sizeof(Scheme_Quick_Env) * total);
  pos = 0;
  while (pos < total) {
    memcpy(cont->saved_env_stack + pos, 
	   env_stack->array + env_stack->stack_pos,
	   (NUM_QUICK_ENVS_PER_SEGMENT - env_stack->stack_pos)
	   * sizeof(Scheme_Quick_Env));
    pos += NUM_QUICK_ENVS_PER_SEGMENT;
    env_stack = env_stack->next;
  }
#endif

  if (scheme_setjmpup(&cont->buf, scheme_current_process->cc_start)) {
    /* Do dynamic winds */
    if (cont->dw) {
      cont->dw->next = start = NULL;
      for (dw = cont->dw; dw != cont->common; dw = dw->prev) {
	if (dw->prev)
	  dw->prev->next = dw;
	else
	  start = dw;
      }
      for (dw = cont->common ? cont->common->next : start; dw; dw = dw->next)
	if (dw->pre)
	  dw->pre(dw->data);
      if (cont->dw && cont->dw != cont->common)
	memcpy(&scheme_error_buf, &cont->dw->saveerr, sizeof(jmp_buf));
    }

    scheme_current_process->stack_overflow = cont->orig_overflow;
    scheme_current_process->overflow = cont->save_overflow;
    scheme_restore_env_stack(cont->envss);

#if MAX_QUICK_VALUES > 0
    env_stack = scheme_current_process->env_stack;
    while (env_stack->prev) {
      env_stack->prev->next = env_stack;
      env_stack = env_stack->prev;
    }

    total = cont->saved_stack_size;
    pos = 0;
    while (pos < total) {
      memcpy(env_stack->array + env_stack->stack_pos,
	     cont->saved_env_stack + pos, 
	     (NUM_QUICK_ENVS_PER_SEGMENT - env_stack->stack_pos)
	     * sizeof(Scheme_Quick_Env));
      pos += NUM_QUICK_ENVS_PER_SEGMENT;
    }
#endif

    return cont->value;
  } else {
    p[0] = (Scheme_Object *)cont;
    ret = _scheme_tail_apply(argv[0], 1, p);
    return ret;
  }
}

typedef struct {
  Scheme_Object *pre, *act, *post;
} Dyn_Wind;

static void pre_dyn_wind(void *d)
{
  Dyn_Wind *dw;
  dw = (Dyn_Wind *)d;

  _scheme_apply(dw->pre, 0, NULL);
}

static Scheme_Object *do_dyn_wind(void *d)
{
  Dyn_Wind *dw;
  dw = (Dyn_Wind *)d;

  return _scheme_apply(dw->act, 0, NULL);
}

static void post_dyn_wind(void *d)
{
  Dyn_Wind *dw;
  dw = (Dyn_Wind *)d;

  (void)_scheme_apply(dw->post, 0, NULL);
}

static Scheme_Object *dynamic_wind(int c, Scheme_Object *p[])
{
  Dyn_Wind *dw;

  if (c != 3)
    scheme_wrong_count("dynamic-wind", 3, 3, c, p);

  dw = (Dyn_Wind *)scheme_malloc(sizeof(Dyn_Wind));

  dw->pre = p[0];
  dw->act = p[1];
  dw->post = p[2];

  return scheme_dynamic_wind(pre_dyn_wind, do_dyn_wind, post_dyn_wind, NULL,
			     (void *)dw);
}
 
Scheme_Object *scheme_dynamic_wind(void (*pre)(void *),
				   Scheme_Object *(*act)(void *),
				   void (*post)(void *), 
				   Scheme_Object *(*jmp_handler)(void *),
				   void *data)
{
  Scheme_Object *v;
  int err;
  Scheme_Dynamic_Wind *dw;

  dw = (Scheme_Dynamic_Wind *)scheme_malloc(sizeof(Scheme_Dynamic_Wind));

  dw->data = data;
  dw->pre = pre;
  dw->post = post;
  dw->prev = scheme_current_process->dw;

  scheme_current_process->dw = dw;

  memcpy(&dw->saveerr, &scheme_error_buf, sizeof(jmp_buf));

  scheme_save_env_stack(dw->envss);

  if (pre)
    pre(data);

  if (setjmp(scheme_error_buf)) {
    scheme_restore_env_stack(dw->envss);
    if (jmp_handler)
      v = jmp_handler(data);
    else
      v = NULL;
    err = !v;
  } else {
    v = act(data);
    err = 0;
  }

  if (post) {
    if (setjmp(scheme_error_buf)) {
      scheme_restore_env_stack(dw->envss);
      err = 1;
    } else {
      post(data);
    }
  }

  if (err)
    scheme_longjmp(dw->saveerr, 1);

  memcpy(&scheme_error_buf, &dw->saveerr, sizeof(jmp_buf));

  scheme_current_process->dw = dw->prev;

  return v;
}

static Scheme_Object *
void_func (int argc, Scheme_Object *argv[])
{
  return scheme_void;
}

static Scheme_Object *
is_void_func (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("void?", 1, 1, argc, argv);
  return SAME_OBJ(argv[0], scheme_void) ? scheme_true : scheme_false;
}

#ifdef TIME_SYNTAX

static Scheme_Object *time_apply(int argc, Scheme_Object *argv[])
{
#ifdef USE_FTIME
  struct timeb start, end;
#else
  struct timeval start, end;
#endif
  long cpustart, cpuend;
  long dur, cpudur;
  Scheme_Object *v;

  if (argc != 1)
    scheme_wrong_count("time-apply", 1, 1, argc, argv);
  if (!SCHEME_PROCP(argv[0]))
    scheme_wrong_type("time-apply", "procedure", 0, argc, argv);

#ifdef USE_FTIME
  ftime(&start);
#else
  gettimeofday(&start, NULL);
#endif
  cpustart = clock();
  v = _scheme_apply(argv[0], 0, NULL);
  cpuend = clock();
#ifdef USE_FTIME
  ftime(&end);
#else
  gettimeofday(&end, NULL);
#endif

#ifdef USE_FTIME
  dur = - start.time * 1000 + end.time * 1000 
    - start.millitm + end.millitm;
#else
  dur = - start.tv_sec * 1000 + end.tv_sec * 1000 
    - start.tv_usec / 1000 + end.tv_usec / 1000;
#endif
  cpudur = (cpuend - cpustart) / 1000;

  scheme_display(scheme_make_string("cpu time: "), scheme_stdout_port);
  scheme_display(scheme_make_integer(cpudur), scheme_stdout_port);
  scheme_display(scheme_make_string("  real time: "), scheme_stdout_port);
  scheme_display(scheme_make_integer(dur), scheme_stdout_port);
  scheme_display(scheme_make_string("\n"), scheme_stdout_port);

  return v;
}

#endif

/****************************************************************/

void scheme_rep()
{
  (void)scheme_apply(rep_prim, 0, NULL);
}

static Scheme_Object *do_rep(void *v)
{
  Scheme_Object *obj;

  scheme_display(scheme_make_string("> "), scheme_stdout_port);
  scheme_flush_output(scheme_stdout_port);
    
  obj = scheme_read(scheme_stdin_port);
  if (SCHEME_EOFP(obj))
    return obj;

  obj = scheme_evaluator(obj);

  if (!SAME_OBJ(obj, scheme_void)) {
    scheme_write (obj, scheme_stdout_port);
    scheme_display (scheme_make_string("\n"), scheme_stdout_port);
  }

  return scheme_void;
}

static Scheme_Object *handle_rep(void *v)
{
  if (scheme_jumping_to_continuation)
    return NULL;

  /* There was an error. */
  scheme_current_process->stack_overflow = *(int *)v;
  return scheme_void;
}

static Scheme_Object *rep(int argc, Scheme_Object *argv[])
{
  Scheme_Object *v;
  int *orig_overflow_ptr;

  if (argc)
    scheme_wrong_count("read-eval-print-loop", 0, 0, argc, argv);

  orig_overflow_ptr = (int *)scheme_malloc_atomic(sizeof(int));
  *orig_overflow_ptr = scheme_current_process->stack_overflow;

  do {
    v = scheme_dynamic_wind(NULL, do_rep, NULL, handle_rep, 
			    (void *)orig_overflow_ptr);
    if (SCHEME_EOFP(v))
      return scheme_void;
  } while(1);
}

/****************************************************************/

#define BOOL(x) (x ? scheme_true : scheme_false)

static Scheme_Object *write_compiled_closure(Scheme_Object *obj)
{
  Scheme_Closure_Compilation_Data *data;

  data = (Scheme_Closure_Compilation_Data *)obj;

  return CONS(BOOL(data->must_allocate),
	      CONS(BOOL(data->has_rest),
		   CONS(scheme_make_integer(data->num_params),
			CONS(scheme_make_integer(data->extra_frame_slots),
			     data->code))));
}

static Scheme_Object *read_compiled_closure(Scheme_Object *obj)
{
  Scheme_Closure_Compilation_Data *data;
  Scheme_Object *v;

#define BAD_CC "bad compiled closure"

  data  = (Scheme_Closure_Compilation_Data *)scheme_malloc_stubborn(sizeof(Scheme_Closure_Compilation_Data));

  data->type = scheme_closure_compilation_type;

  SCHEME_ASSERT(SCHEME_PAIRP(obj), BAD_CC);
  v = SCHEME_CAR(obj);
  obj = SCHEME_CDR(obj);
  SCHEME_ASSERT(SCHEME_BOOLP(v), BAD_CC);
  data->must_allocate = SAME_OBJ(v, scheme_true);

  SCHEME_ASSERT(SCHEME_PAIRP(obj), BAD_CC);
  v = SCHEME_CAR(obj);
  obj = SCHEME_CDR(obj);
  SCHEME_ASSERT(SCHEME_BOOLP(v), BAD_CC);
  data->has_rest = SAME_OBJ(v, scheme_true);

  SCHEME_ASSERT(SCHEME_PAIRP(obj), BAD_CC);
  v = SCHEME_CAR(obj);
  obj = SCHEME_CDR(obj);
  SCHEME_ASSERT(SCHEME_INTP(v), BAD_CC);
  data->num_params = SCHEME_INT_VAL(v);

  SCHEME_ASSERT(SCHEME_PAIRP(obj), BAD_CC);
  v = SCHEME_CAR(obj);
  obj = SCHEME_CDR(obj);
  SCHEME_ASSERT(SCHEME_INTP(v), BAD_CC);
  data->extra_frame_slots = SCHEME_INT_VAL(v);

  data->code = obj;

  scheme_end_stubborn_change((void *)data);

  return (Scheme_Object *)data;
}

