/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"

/* Quasiquote is difficult to implement in Scheme without cond, etc., which
   are implemented using quasiquote. So, the old primitive implementation
   can be turned on here: */
#define QUASIQUOTE_ON 0

/* globals */
Scheme_Object *scheme_define_syntax, *scheme_defmacro_syntax;

/* locals */
static Scheme_Object *lambda_syntax(Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec);
static Scheme_Object *lambda_expand(Scheme_Object *form, Scheme_Env *env, int depth);
static Scheme_Object *define_syntax(Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec);
static Scheme_Object *define_expand(Scheme_Object *form, Scheme_Env *env, int depth);
static Scheme_Object *quote_syntax(Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec);
static Scheme_Object *quote_expand(Scheme_Object *form, Scheme_Env *env, int depth);
static Scheme_Object *if_syntax(Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec);
static Scheme_Object *set_expand(Scheme_Object *form, Scheme_Env *env, int depth);
static Scheme_Object *set_syntax(Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec);
static Scheme_Object *case_lambda_syntax(Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec);
static Scheme_Object *case_lambda_expand(Scheme_Object *form, Scheme_Env *env, int depth);
static Scheme_Object *let_expand(Scheme_Object *form, Scheme_Env *env, int depth);
static Scheme_Object *let_syntax(Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec);
static Scheme_Object *let_star_syntax (Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec);
static Scheme_Object *letrec_syntax (Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec);
static Scheme_Object *letrec_star_syntax (Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec);
static Scheme_Object *begin_syntax (Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec);
static Scheme_Object *begin_expand (Scheme_Object *form, Scheme_Env *env, int depth);
static Scheme_Object *delay_syntax (Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec);

#if QUASIQUOTE_ON
static Scheme_Object *quasiquote_syntax (Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec);
static Scheme_Object *quasiquote_expand(Scheme_Object *form, Scheme_Env *env, int depth);
#endif

/* non-standard */
static Scheme_Object *defmacro_syntax(Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec);
static Scheme_Object *defmacro_expand(Scheme_Object *form, Scheme_Env *env, int depth);
static Scheme_Object *letmacro_syntax(Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec);
static Scheme_Object *letmacro_expand(Scheme_Object *form, Scheme_Env *env, int depth);

static Scheme_Object *define_execute (Scheme_Object *data, Scheme_Env *env);
static Scheme_Object *quote_execute (Scheme_Object *data, Scheme_Env *env);
static Scheme_Object *set_execute (Scheme_Object *data, Scheme_Env *env);
static Scheme_Object *gen_let_execute (Scheme_Object *data, Scheme_Env *env);
static Scheme_Object *delay_execute (Scheme_Object *expr, Scheme_Env *env);
#if QUASIQUOTE_ON
static Scheme_Object *quasi_execute(Scheme_Object *expr, Scheme_Env *env);
#endif
static Scheme_Object *defmacro_execute(Scheme_Object *expr, Scheme_Env *env);
static Scheme_Object *case_lambda_execute(Scheme_Object *expr, Scheme_Env *env);
static Scheme_Object *void_execute(Scheme_Object *expr, Scheme_Env *env);

static Scheme_Object *write_let(Scheme_Object *obj);
static Scheme_Object *read_let(Scheme_Object *obj);
static Scheme_Object *write_let_value(Scheme_Object *obj);
static Scheme_Object *read_let_value(Scheme_Object *obj);
static Scheme_Object *write_let_void(Scheme_Object *obj);
static Scheme_Object *read_let_void(Scheme_Object *obj);
static Scheme_Object *write_uq(Scheme_Object *obj);
static Scheme_Object *read_uq(Scheme_Object *obj);
static Scheme_Object *write_uq_splice(Scheme_Object *obj);
static Scheme_Object *read_uq_splice(Scheme_Object *obj);
static Scheme_Object *write_quoted_vector(Scheme_Object *obj);
static Scheme_Object *read_quoted_vector(Scheme_Object *obj);

/* symbols */
#ifdef  QUASIQUOTE_ON
static Scheme_Object *scheme_quasiquote;
static Scheme_Object *scheme_unquote;
static Scheme_Object *scheme_unquote_splicing;
#endif
static Scheme_Object *scheme_define;
static Scheme_Object *scheme_lambda;
static Scheme_Object *scheme_letrec;
static Scheme_Object *scheme_letrec_star;
static Scheme_Object *scheme_let_star;
static Scheme_Object *scheme_let;
static Scheme_Object *scheme_begin;
static Scheme_Object *scheme_else;
static Scheme_Object *scheme_arrow;
static Scheme_Object *dummy_symbol;

#define cons(a,b) scheme_make_pair(a,b)

#define max(a, b) (((a) > (b)) ? (a) : (b))

void 
scheme_init_syntax (Scheme_Env *env)
{
  if (scheme_starting_up) {
#ifdef QUASIQUOTE_ON
    scheme_quasiquote = scheme_intern_symbol ("quasiquote");
    scheme_unquote = scheme_intern_symbol ("unquote");
    scheme_unquote_splicing = scheme_intern_symbol ("unquote-splicing");
#endif
    scheme_define = scheme_intern_symbol ("define");
    scheme_lambda = scheme_intern_symbol ("lambda");
    scheme_letrec = scheme_intern_symbol ("letrec");
    scheme_letrec_star = scheme_intern_symbol ("letrec*");
    scheme_let_star = scheme_intern_symbol ("let*");
    scheme_let = scheme_intern_symbol ("let");
    scheme_begin = scheme_intern_symbol ("begin");
    scheme_else = scheme_intern_symbol ("else");
    scheme_arrow = scheme_intern_symbol ("=>");
    dummy_symbol = scheme_make_symbol("dummy"); /* uninterned */

    scheme_register_syntax_executer("-d", define_execute);
    scheme_register_syntax_executer("-!", set_execute);
    scheme_register_syntax_executer("-q", quote_execute);
    scheme_register_syntax_executer("-t", gen_let_execute);
    scheme_register_syntax_executer("-y", delay_execute);
#if QUASIQUOTE_ON
    scheme_register_syntax_executer("-qq", quasi_execute);
#endif
    scheme_register_syntax_executer("-dm", defmacro_execute);
    scheme_register_syntax_executer("-cl", case_lambda_execute);
    scheme_register_syntax_executer("-v", void_execute);

    scheme_install_type_writer(scheme_let_data_type, write_let);
    scheme_install_type_reader(scheme_let_data_type, read_let);
    scheme_install_type_writer(scheme_let_value_compilation_type, write_let_value);
    scheme_install_type_reader(scheme_let_value_compilation_type, read_let_value);
    scheme_install_type_writer(scheme_let_void_compilation_type, write_let_void);
    scheme_install_type_reader(scheme_let_void_compilation_type, read_let_void);

    scheme_install_type_writer(scheme_compiled_unquote_type, write_uq);
    scheme_install_type_reader(scheme_compiled_unquote_type, read_uq);
    scheme_install_type_writer(scheme_compiled_unquote_splicing_type, 
			       write_uq_splice);
    scheme_install_type_reader(scheme_compiled_unquote_splicing_type, 
			       read_uq_splice);
    scheme_install_type_writer(scheme_compiled_quoted_vector_type, 
			       write_quoted_vector);
    scheme_install_type_reader(scheme_compiled_quoted_vector_type, 
			       read_quoted_vector);

    scheme_define_syntax = scheme_make_compiled_syntax(define_syntax, 
						       define_expand);
    scheme_defmacro_syntax = scheme_make_compiled_syntax(defmacro_syntax, 
							 defmacro_expand);
  }

  scheme_add_global_keyword ("lambda", scheme_make_compiled_syntax (lambda_syntax, lambda_expand), env);
  scheme_add_global_keyword ("define", scheme_define_syntax, env);
  scheme_add_global_keyword ("quote", scheme_make_compiled_syntax (quote_syntax, quote_expand), env);
  scheme_add_global_keyword ("if", scheme_make_compiled_syntax (if_syntax, scheme_expand_simple), env);
  scheme_add_global_keyword ("set!", scheme_make_compiled_syntax (set_syntax, set_expand), env);

  scheme_add_global_keyword ("case-lambda", scheme_make_compiled_syntax (case_lambda_syntax, case_lambda_expand), env);
  scheme_add_global_keyword ("let", scheme_make_compiled_syntax (let_syntax, let_expand), env);
  scheme_add_global_constant ("let*", scheme_make_compiled_syntax (let_star_syntax, let_expand), env);
  scheme_add_global_keyword ("letrec", scheme_make_compiled_syntax (letrec_syntax, let_expand), env);  
  scheme_add_global_keyword ("letrec*", scheme_make_compiled_syntax (letrec_star_syntax, let_expand), env);

  scheme_add_global_keyword ("begin", scheme_make_compiled_syntax (begin_syntax, begin_expand), env);

  scheme_add_global_keyword ("delay", scheme_make_compiled_syntax (delay_syntax, scheme_expand_simple), env);
#if QUASIQUOTE_ON
  scheme_add_global_keyword ("quasiquote", scheme_make_compiled_syntax (quasiquote_syntax, quasiquote_expand), env);
#endif
  scheme_add_global_keyword ("defmacro", scheme_defmacro_syntax, env);
  scheme_add_global_keyword ("letmacro", scheme_make_compiled_syntax (letmacro_syntax, letmacro_expand), env);
}

Scheme_Object *
scheme_make_compiled_syntax(Scheme_Syntax *proc, 
			    Scheme_Syntax_Expander *eproc)
{
  Scheme_Object *syntax;

  syntax = scheme_alloc_eternal_object ();
  syntax->type = scheme_compiled_syntax_type;
  SCHEME_SYNTAX(syntax) = proc;
  SCHEME_SYNTAX_EXP(syntax) = eproc;
  return (syntax);
}

static int check_form(char *name, Scheme_Object *form)
{
  int i;

  for (i = 0; SCHEME_PAIRP(form); i++)
    form = SCHEME_CDR(form);

  if (!SCHEME_NULLP(form))
    scheme_signal_error("%s: bad form (improper list)", name);

  return i;
}

static void bad_form(char *name, int l)
{ 
  scheme_signal_error("%s: bad form (has %d part%s)", name, l - 1,
		      (l != 2) ? "s" : "");
}

/* builtin syntax */

static void lambda_check(Scheme_Object *form)
{
  SCHEME_ASSERT (SCHEME_PAIRP(form), "badly formed lambda");
  SCHEME_ASSERT (SCHEME_PAIRP(SCHEME_CDR(form)), "badly formed lambda");
  SCHEME_ASSERT (SCHEME_PAIRP(SCHEME_CDDR(form)), "badly formed lambda");
}

static Scheme_Object *
lambda_syntax (Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec)
{
  lambda_check(form);
  return (scheme_make_closure_compilation (env, SCHEME_CDR(form), rec));
}

static Scheme_Object *
lambda_expand(Scheme_Object *form, Scheme_Env *env, int depth)
{
  Scheme_Object *args;

  lambda_check(form);
  
  args = SCHEME_CAR(SCHEME_CDR(form));

  env = scheme_add_frame(args, env, (depth < 0) ? SCHEME_ANNOTATING : 0);

  if (depth < 0) {
    Scheme_Object *first = scheme_null, *last = NULL, *arg;
    while (!SCHEME_NULLP(args)) {
      if (SCHEME_PAIRP(args)) {
	arg = SCHEME_CAR(args);
	args = SCHEME_CDR(args);

	arg = scheme_make_pair(scheme_annotate_frame_id(env, arg), scheme_null);
      } else {
	arg = scheme_annotate_frame_id(env, args);
	args = scheme_null;
      }


      if (!last)
	first = arg;
      else
	SCHEME_CDR(last) = arg;
      last = arg;
    }

    args = first;
  }

  return cons(SCHEME_CAR(form),
	      cons(args,
		   scheme_expand_block(SCHEME_CDR(SCHEME_CDR(form)),
				       env,
				       depth)));
}

static Scheme_Object *
define_execute (Scheme_Object *data, Scheme_Env *env)
{
  Scheme_Object *var, *val;

  var = SCHEME_CAR(data);
  val = SCHEME_CDR(data);

  val = _scheme_eval_compiled(val, env);

  scheme_add_global_symbol(var, val, env);
  return scheme_void;
}

static void define_parse(Scheme_Object *form, 
			 Scheme_Object **var, Scheme_Object **val,
			 Scheme_Env *env)
{
  Scheme_Object *sec;

  SCHEME_ASSERT(SCHEME_PAIRP(SCHEME_CDR(form)),
		"define: bad form");

  sec = SCHEME_CAR (SCHEME_CDR (form));

  if (SCHEME_PAIRP(sec)) {
    Scheme_Object *forms, *params;

    *var = SCHEME_CAR (sec);
    params = SCHEME_CDR (sec);
    forms = SCHEME_CDR (SCHEME_CDR (form));
    *val = cons(scheme_lambda, scheme_make_pair (params, forms));
  } else if (SCHEME_SYMBOLP(sec)) {
    int len;

    len = check_form("define", form);
    if (len != 3)
      bad_form("define", len);
    *var = sec;
    *val = SCHEME_CAR (SCHEME_CDR (SCHEME_CDR (form)));
  } else {
    scheme_signal_error ("define: second arg must be symbol or list");
    *var = *val = NULL;
  }

  scheme_check_identifier("define", *var, NULL, env);
}

static Scheme_Object *
define_syntax (Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec)
{
  Scheme_Object *var, *val;

  if (env->basic.next)
    scheme_signal_error ("define: illegal use");

  define_parse(form, &var, &val, env);
  
  rec->max_let_depth = 0;
  rec->can_return_closure = 1;

  val = scheme_compile_expr(val, env, rec);

  return scheme_make_syntax_compilation(define_execute,
					cons(var, val));
}

static Scheme_Object *
define_expand(Scheme_Object *form, Scheme_Env *env, int depth)
{
  Scheme_Object *var, *val;

  define_parse(form, &var, &val, env);

  return cons(SCHEME_CAR(form),
	      cons(var,
		   cons(scheme_expand_expr(val, env, depth), scheme_null)));
}

static Scheme_Object *
quote_execute (Scheme_Object *var, Scheme_Env *env)
{
  return var;
}

static Scheme_Object *
quote_syntax (Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec)
{
  Scheme_Object *v;

  SCHEME_ASSERT (SCHEME_PAIRP(SCHEME_CDR(form))
		 && SCHEME_NULLP(SCHEME_CDR(SCHEME_CDR(form))),
		 "quote: wrong number of args");

  rec->max_let_depth = 0;
  rec->can_return_closure = 0;

  
  v = SCHEME_CAR(SCHEME_CDR(form));

  /* If this is at the top level, we may have to explicitly quote it,
     even in compiled form, for the read-eval-print loop's sake. */
  if (rec->can_optimize_constants || env->basic.next)
    return v;
  else
    return scheme_make_syntax_compilation(quote_execute, v);
}

static Scheme_Object *
quote_expand(Scheme_Object *form, Scheme_Env *env, int depth)
{
  return form;
}

static Scheme_Object *
if_syntax (Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec)
{
  int len;
  Scheme_Object *test, *thenp, *elsep;
  Scheme_Compile_Info recs[3];

  len = check_form("if", form);
  if (!(((len == 3) || (len == 4))))
    bad_form("if", len);

  test = SCHEME_CAR(SCHEME_CDR (form));
  thenp = SCHEME_CAR (SCHEME_CDR (SCHEME_CDR (form)));
  if (len == 4)
    elsep = SCHEME_CAR (SCHEME_CDR (SCHEME_CDR ((SCHEME_CDR (form)))));
  else
    elsep = scheme_compiled_void(rec->can_optimize_constants);

  scheme_init_compile_recs(rec, recs, 3);

  test = scheme_compile_expr(test, env, &recs[0]);
  thenp = scheme_compile_expr(thenp, env, &recs[1]);
  elsep = scheme_compile_expr(elsep, env, &recs[2]);

  scheme_merge_compile_recs(rec, recs, 3);

  return scheme_make_branch_compilation(test, thenp, elsep);
}

static Scheme_Object *
set_execute (Scheme_Object *data, Scheme_Env *env)
{
  Scheme_Object *var, *val;

  var = SCHEME_CAR(data);
  val = _scheme_eval_compiled(SCHEME_CDR(data), env);

  scheme_set_value(var, val, env);

  return scheme_void;
}

static Scheme_Object *
set_syntax (Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec)
{
  Scheme_Object *var, *val, *name;
  int l;

  l = check_form("set!", form);
  if (l != 3)
    bad_form("set!", l);

  name = SCHEME_CAR (SCHEME_CDR (form));
  
  scheme_check_identifier("set!", name, NULL, env);

  var = scheme_static_distance(name, env, rec->can_optimize_constants);
  val = scheme_compile_expr(SCHEME_CAR (SCHEME_CDR (SCHEME_CDR (form))), 
			    env, rec);
  
  if (SAME_TYPE(SCHEME_TYPE(var), scheme_static_distance_type)
      || SAME_TYPE(SCHEME_TYPE(var), scheme_static_distance_unbox_type)) {
    int j = SCHEME_SDFRAME_VAL(var);
    while (j--)
      env = env->basic.next;
    env->basic.flags |= SCHEME_HAS_SET_BANG;
  } else if (NOT_SAME_TYPE(SCHEME_TYPE(var), scheme_variable_type))
    scheme_signal_error("set!: cannot modifiy value of constant %s",
			SCHEME_SYM_VAL(name));


  return scheme_make_syntax_compilation(set_execute, cons(var, val));
}

static Scheme_Object *
set_expand(Scheme_Object *form, Scheme_Env *env, int depth)
{
  SCHEME_ASSERT(SCHEME_PAIRP(form), "bad set! form");

  return cons(SCHEME_CAR(form),
	      scheme_expand_simple(SCHEME_CDR(form), env, depth));
}

static Scheme_Object *
case_lambda_execute(Scheme_Object *expr, Scheme_Env *env)
{
  Scheme_Sequence *seqin, *seqout;
  int i;

  seqin = (Scheme_Sequence *)expr;

  seqout = 
    (Scheme_Sequence *)scheme_malloc_stubborn(sizeof(Scheme_Sequence)
					      + ((seqin->count - 1)
						 * sizeof(Scheme_Object *)));

  seqout->type = scheme_case_closure_type;
  seqout->count = seqin->count;

  for (i = 0; i < seqin->count; i++) {
    if (SAME_TYPE(SCHEME_TYPE(seqin->array[i]), scheme_compiled_closure_type))
      seqout->array[i] = seqin->array[i];
    else
      seqout->array[i] = scheme_make_compiled_closure(env, seqin->array[i]);
  }

  scheme_end_stubborn_change((void *)seqout);

  return (Scheme_Object *)seqout;
}

static Scheme_Object *
case_lambda_syntax (Scheme_Object *form, Scheme_Env *env, 
		     Scheme_Compile_Info *rec)
{
  Scheme_Object *list, *last, *c;
  
  form = SCHEME_CDR(form);
  if (!SCHEME_PAIRP(form))
    scheme_signal_error("bad `case-lambda' form");
  if (SCHEME_NULLP(SCHEME_CDR(form)))
    return lambda_syntax(scheme_make_pair(scheme_lambda,
					  SCHEME_CAR(form)), env, rec);

  list = last = NULL;
  while (SCHEME_PAIRP(form)) {
    c = scheme_make_pair(scheme_make_pair(scheme_lambda, SCHEME_CAR(form)),
			 scheme_null);
    if (list)
      SCHEME_CDR(last) = c;
    else
      list = c;

    last = c;
    form = SCHEME_CDR(form);
  }

  if (!SCHEME_NULLP(form))
    scheme_signal_error("bad `case-lambda' form");

  form = scheme_make_sequence_compilation(scheme_compile_list(list, env, rec));

  return scheme_make_syntax_compilation(case_lambda_execute, form);
}

static Scheme_Object *
case_lambda_expand(Scheme_Object *form, Scheme_Env *env, int depth)
{
  Scheme_Object *first, *last, *args, *body, *c;

  last = first = cons(SCHEME_CAR(form), scheme_null);
  form = SCHEME_CDR(form);

  while (SCHEME_PAIRP(form)) {
    args = SCHEME_CAR(form);
    if (!SCHEME_PAIRP(args))
      scheme_signal_error("bad `case-lambda' form");

    body = SCHEME_CDR(args);
    args = SCHEME_CAR(args);

    c = cons(cons(args, scheme_expand_block(body, 
					    scheme_add_frame(args, env, 0), 
					    depth)),
	     scheme_null);

    SCHEME_CDR(last) = c;
    last = c;

    form = SCHEME_CDR(form);
  }

  if (!SCHEME_NULLP(form))
    scheme_signal_error("bad `case-lambda' form");

  return first;
}

static Scheme_Object *named_let_syntax(Scheme_Object *form, Scheme_Env *env, 
				       Scheme_Compile_Info *rec, int depth);

typedef struct {
  Scheme_Type type;
  short num_bindings;
  Scheme_Object *body;
} Let_Data;

static Scheme_Object *
gen_let_execute (Scheme_Object *data, Scheme_Env *env)
{
  Scheme_Env *frame;
  Let_Data *let;

  let = (Let_Data *)data;

  frame = scheme_new_frame(let->num_bindings);
  env = scheme_extend_env(frame, env);

  return _scheme_tail_eval(let->body, env);
}

static Scheme_Object *
gen_let_syntax (Scheme_Object *form, Scheme_Env *env, char *formname,
		int star, int recursive, Scheme_Compile_Info *rec)
{
  Scheme_Object *bindings, *binding, *name, **names, *forms;
  int new_frame, bind_start;
  int num_bindings, i, j;
  Scheme_Env *frame, *orig_env;
  Scheme_Compile_Info *recs;
  Scheme_Object *first = NULL;
  Scheme_Let_Value *last = NULL, *lv;

  if (scheme_list_length(form) < 3)
    scheme_signal_error("badly formed `%s' form", formname);

  bindings = SCHEME_CAR (SCHEME_CDR (form));
  num_bindings = scheme_list_length (bindings);

  forms = SCHEME_CDR (SCHEME_CDR (form));

  if (!num_bindings) {
    if (!env->basic.next) {
      /* Hack: adds a useless frame to block illegal
	 (define ...) statements */
      frame = scheme_new_frame(0);
      env = scheme_extend_env(frame, env);
    }
    return scheme_compile_sequence(forms, env, rec);
  }

  names = (Scheme_Object **)scheme_malloc_stubborn(num_bindings
						   * sizeof(Scheme_Object*));

  if (rec->base_let_depth < 0) {
    new_frame = 1;
    bind_start = 0;
    rec->base_let_depth = num_bindings;
  } else {
    new_frame = 0;
    bind_start = rec->base_let_depth;
    rec->base_let_depth += num_bindings;
  }

  if (new_frame) {
    orig_env = NULL;
    frame = scheme_new_compilation_frame(num_bindings + bind_start, 0);
    env = scheme_extend_env(frame, env);
  } else {
    orig_env = env;
    env = frame = scheme_lengthen_compilation_frame(num_bindings, env);
  }

  for (i = num_bindings; i--; ) 
    scheme_add_compilation_binding (i + bind_start, dummy_symbol, frame);

  recs = (Scheme_Compile_Info *)scheme_malloc((num_bindings + 1)
					      * sizeof(Scheme_Compile_Info));

  scheme_init_compile_recs(rec, recs, num_bindings + 1);

  /* Values of let-bound variables can't share the frame... */
  for (i = 0; i < num_bindings; i++)
    recs[i].base_let_depth = -1;

  for (i = 0; i < num_bindings; i++) {
    if (!SCHEME_PAIRP(bindings))
      scheme_signal_error("badly formed `%s' form", formname);
    binding = SCHEME_CAR (bindings);
    if (!(SCHEME_PAIRP(binding) && SCHEME_PAIRP(SCHEME_CDR(binding))
	  && SCHEME_NULLP(SCHEME_CDDR(binding))))
      scheme_signal_error("badly formed `%s' form", formname);
    
    name = SCHEME_CAR(binding);
    scheme_check_identifier(formname, name, NULL, env);

    if (!star)
      for (j = 0; j < i ; j++)
	if (SAME_OBJ(name, names[j]))
	  scheme_signal_error("multiple bindings of \"%s\" in "
			      "`%s' form", SCHEME_SYM_VAL(name), formname);

    lv = (Scheme_Let_Value *)scheme_malloc(sizeof(Scheme_Let_Value));
    lv->type = scheme_let_value_compilation_type;
    if (!last)
      first = (Scheme_Object *)lv;
    else
      last->body = (Scheme_Object *)lv;
    last = lv;
    lv->position = i + bind_start;

    names[i] = name;
    if (!recursive)
      lv->value = scheme_compile_expr(SCHEME_CADR(binding), env, &recs[i]);
    else
      lv->value = SCHEME_CADR(binding);
    
    if (star || recursive)
      scheme_add_compilation_binding(i + bind_start, name, frame);
    
    bindings = SCHEME_CDR (bindings);
  }
  
  if (!star && !recursive)
    for (i = 0; i < num_bindings; i++)
      scheme_add_compilation_binding(i + bind_start, names[i], frame);

  if (recursive) {
    Scheme_Let_Void *lvoid;

    lv = (Scheme_Let_Value *)first;
    for (i = 0; i < num_bindings; i++, lv = (Scheme_Let_Value *)lv->body)
      lv->value = scheme_compile_expr(lv->value, env, &recs[i]);

    lvoid = (Scheme_Let_Void *)scheme_malloc(sizeof(Scheme_Let_Void));
    lvoid->type = scheme_let_void_compilation_type;
    lvoid->body = first;
    lvoid->start = bind_start;
    lvoid->end = bind_start + num_bindings;

    first = (Scheme_Object *)lvoid;
  }

  last->body = scheme_compile_sequence(forms, env, recs + num_bindings);
  
  scheme_merge_compile_recs(rec, recs, num_bindings + 1);

  scheme_end_stubborn_change((void *)names);

  if (orig_env)
    scheme_frame_backinfo(orig_env, env);

  if (new_frame) {
    Let_Data *let;

    let = (Let_Data *)scheme_malloc_stubborn(sizeof(Let_Data));
    
    let->type = scheme_let_data_type;

    let->body = first;
    let->num_bindings = num_bindings + rec->max_let_depth;

    scheme_end_stubborn_change((void *)let);

    first = scheme_make_syntax_compilation(gen_let_execute, (Scheme_Object *)let);
  }

  rec->max_let_depth += num_bindings;

  return first;
}

static Scheme_Object *
let_expand(Scheme_Object *form, Scheme_Env *env, int depth)
{
  int named, letrec, letstar;
  Scheme_Object *formname, *vars, *body, *first, *last, *name, *v, *vs, *vlist;
  Scheme_Env *use_env;

  formname = SCHEME_CAR(form);

  letrec = (SAME_OBJ(formname, scheme_letrec)
	    || SAME_OBJ(formname, scheme_letrec_star));
  letstar = SAME_OBJ(formname, scheme_let_star);
  named = !letrec && !letstar && SCHEME_SYMBOLP(SCHEME_CAR(SCHEME_CDR(form)));
  
  if (named)
    return named_let_syntax(form, env, NULL, depth);

  vars = SCHEME_CDR(form);

  if (!SCHEME_PAIRP(vars))
    scheme_signal_error("bad `%s' form", SCHEME_SYM_VAL(formname));
  body = SCHEME_CDR(vars);
  vars = SCHEME_CAR(vars);

  if (letstar) {
    if (!SCHEME_NULLP(vars)) {
      body = cons(scheme_let,
		  cons(cons(SCHEME_CAR(vars), scheme_null),
		       cons(cons(scheme_let_star, 
				 cons(SCHEME_CDR(vars), body)),
			    scheme_null)));
    } else
      body = cons(scheme_let, cons(scheme_null, body));

    --depth;
    if (!depth)
      return body;
    else
      return scheme_expand_expr(body, env, depth);
  }

  vlist = scheme_null;
  vs = vars;
  while (SCHEME_PAIRP(vs)) {
    v = SCHEME_CAR(vs);
    if (!SCHEME_PAIRP(v))
      scheme_signal_error("bad `%s' form", SCHEME_SYM_VAL(formname));

    name = SCHEME_CAR(v);
    
    scheme_check_identifier(SCHEME_SYM_VAL(formname), name, NULL, env);

    vlist = cons(name, vlist);

    vs = SCHEME_CDR(vs);
  }

  use_env = env;
  env = scheme_add_frame(vlist, env, (depth < 0) ? SCHEME_ANNOTATING : 0);

  if (letrec)
    use_env = env;

  first = last = NULL;
  while (SCHEME_PAIRP(vars)) {
    v = SCHEME_CAR(vars);

    name = SCHEME_CAR(v);
    if (depth < 0)
      name = scheme_annotate_frame_id(env, name);

    v = scheme_expand_expr(SCHEME_CDR(v), use_env, depth);
    v = cons(cons(name, v), scheme_null);

    if (!first)
      first = v;
    else
      SCHEME_CDR(last) = v;

    last = v;

    vars = SCHEME_CDR(vars);
  }

  if (!SCHEME_NULLP(vars))
    scheme_signal_error("bad `%s' form", SCHEME_SYM_VAL(formname));

  if (!first)
    first = scheme_null;

  body = scheme_expand_block(body, env, depth);

  if (named)
    return cons(SCHEME_CAR(form),
		cons(SCHEME_CADR(form),
		     cons(first, body)));
  else
    return cons(SCHEME_CAR(form),
		cons(first, body));
}

static Scheme_Object *
let_syntax (Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec)
{
  if (!SCHEME_PAIRP(SCHEME_CDR(form)))
    scheme_signal_error("let: bad form (improper list)");
  if (SCHEME_SYMBOLP (SCHEME_CAR (SCHEME_CDR (form))))
    return named_let_syntax (form, env, rec, 0);
  
  return gen_let_syntax(form, env, "let", 0, 0, rec);
}

static Scheme_Object *
let_star_syntax (Scheme_Object *form, Scheme_Env *env, 
		 Scheme_Compile_Info *rec)
{
  return gen_let_syntax(form, env, "let*", 1, 0, rec);
}

static Scheme_Object *
letrec_syntax (Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec)
{
  return gen_let_syntax(form, env, "letrec", 0, 1, rec);
}

static Scheme_Object *
letrec_star_syntax(Scheme_Object *form,Scheme_Env*env,Scheme_Compile_Info *rec)
{
  return gen_let_syntax(form, env, "letrec*", 0, 1, rec);
}


static Scheme_Object *nl_car(Scheme_Object *l)
{
  Scheme_Object *s;
  
  if (!SCHEME_PAIRP(l))
    scheme_signal_error("named let: bad form (non-symbol-value-pair)");
  s = SCHEME_CAR(l);
  if (!SCHEME_SYMBOLP(s))
    scheme_signal_error("named let: bad form (non-symbol variable)");

  return s;
}

static Scheme_Object *nl_cadr(Scheme_Object *l)
{
  if (!SCHEME_PAIRP(l) || !SCHEME_PAIRP(SCHEME_CDR(l)))
    scheme_signal_error("named let: bad form (non-symbol-value-pair)");
  
  if (!SCHEME_NULLP(SCHEME_CDR(SCHEME_CDR(l))))
    scheme_signal_error("named let: bad form "
			"(extra data in variable-value pair)");
  
  return SCHEME_CADR(l);
}

static Scheme_Object *
named_let_syntax (Scheme_Object *form, Scheme_Env *env, 
		  Scheme_Compile_Info *rec, int depth)
{
  Scheme_Object *name, *bindings, *vars, *vals, *forms;
  Scheme_Object *proc, *init, *letrec;

  SCHEME_ASSERT(SCHEME_PAIRP(SCHEME_CDR (form))
		&& SCHEME_PAIRP(SCHEME_CDR(SCHEME_CDR (form)))
		&& SCHEME_PAIRP(SCHEME_CDR(SCHEME_CDR(SCHEME_CDR (form)))), 
		"bad named `let' form");
  name = SCHEME_CAR (SCHEME_CDR (form));
  bindings = SCHEME_CAR (SCHEME_CDR (SCHEME_CDR (form)));
  SCHEME_ASSERT(SCHEME_LISTP(bindings), "bad named `let' form");
  vars = scheme_named_map_1("named let", nl_car, bindings);
  vals = scheme_named_map_1("named let", nl_cadr, bindings);
  
  forms = SCHEME_CDR (SCHEME_CDR (SCHEME_CDR (form)));

  proc = cons(scheme_lambda, cons(vars, forms));
  init = cons(name, vals);
  
  letrec = cons(scheme_letrec,
		cons(cons(cons(name, cons(proc, scheme_null)), scheme_null),
		     cons(init,
			  scheme_null)));

  if (rec)
    return scheme_compile_expr(letrec, env, rec);
  else {
    --depth;
    if (!depth)
      return letrec;
    else
      return scheme_expand_expr(letrec, env, depth);
  }
}

Scheme_Object *scheme_compile_sequence(Scheme_Object *forms,
				       Scheme_Env *env, 
				       Scheme_Compile_Info *rec)
{
  if (SCHEME_PAIRP(forms) && SCHEME_NULLP(SCHEME_CDR(forms)))
    return scheme_compile_expr(SCHEME_CAR(forms), env, rec);
  else
    return scheme_compile_expr(cons(scheme_begin, forms), env, rec);
}

Scheme_Object *scheme_compiled_void(int can_be_value)
{
  if (can_be_value)
    return scheme_void;
  else
    return scheme_make_syntax_compilation(void_execute, scheme_null);
}

static Scheme_Object *
begin_syntax (Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec)
{
  Scheme_Object *forms, *body;

  forms = SCHEME_CDR (form);
  
  if (SCHEME_NULLP(forms))
    return scheme_compiled_void(rec->can_optimize_constants);

  check_form("begin", forms);

  if (SCHEME_NULLP(SCHEME_CDR(forms)))
    return scheme_compile_expr(SCHEME_CAR(forms), env, rec);

  if (env->basic.next)
    /* Not at top-level */
    body = scheme_compile_block(forms, env, rec);
  else
    /* Top level */
    body = scheme_compile_list(forms, env, rec);

  return scheme_make_sequence_compilation(body);
}

static Scheme_Object *
begin_expand (Scheme_Object *form, Scheme_Env *env, int depth)
{
  Scheme_Object *name, *rest;

  check_form("begin", form);

  name = SCHEME_CAR(form);
  rest = SCHEME_CDR(form);

  if (env->basic.next)
    /* Not at top-level: */
    form = cons(name, scheme_expand_block(rest, env, depth));
  else
    /* Top level */
    form =  cons(name, scheme_expand_list(rest, env, depth));

  return form;
}

static Scheme_Object *
delay_execute (Scheme_Object *expr, Scheme_Env *env)
{
  return scheme_make_promise_compiled(expr, env);
}

static Scheme_Object *
delay_syntax (Scheme_Object *form, Scheme_Env *env, Scheme_Compile_Info *rec)
{
  int l;
  l = check_form("delay", form);
  if (l != 2)
    bad_form("delay", l);
  form = scheme_compile_expr(SCHEME_CAR (SCHEME_CDR (form)), env, rec);

  rec->can_return_closure = 1;

  return scheme_make_syntax_compilation(delay_execute, form);
}

#if QUASIQUOTE_ON

static Scheme_Object *quasi_syntax(int edepth, Scheme_Object *x, int level, Scheme_Env *env, Scheme_Compile_Info *rec);

static Scheme_Object *
quasiquote_syntax (Scheme_Object *form, Scheme_Env *env, 
		   Scheme_Compile_Info *rec)
{
  int l;

  l = check_form("quasiquote(`)", form);
  if (l != 2)
    bad_form("quasiquote(`)", l);

  form = quasi_syntax(0, SCHEME_CAR(SCHEME_CDR (form)), 0, env, rec);

  return scheme_make_syntax_compilation(quasi_execute, form);
}

static Scheme_Object *
quasiquote_expand(Scheme_Object *form, Scheme_Env *env, int depth)
{
  int l;

  l = check_form("quasiquote(`)", form);
  if (l != 2)
    bad_form("quasiquote(`)", l);

  return cons(SCHEME_CAR(form),
	      cons(quasi_syntax(depth + 1, SCHEME_CAR(SCHEME_CDR (form)), 
				0, env, NULL), scheme_null));
}

static Scheme_Object *
quasi_execute (Scheme_Object *x, Scheme_Env *env)
{
  Scheme_Object *list, *qcar, *qcdr, *rest;

  if (SAME_TYPE(SCHEME_TYPE(x),  scheme_compiled_quoted_vector_type)) {
    x = quasi_execute((Scheme_Object *)SCHEME_PTR_VAL(x), env);
    return (scheme_list_to_vector (x));
  }

  if (SAME_TYPE(SCHEME_TYPE(x), scheme_compiled_unquote_type))
    return _scheme_eval_compiled(((Scheme_Object *)SCHEME_PTR1_VAL(x)), env);

  if (SAME_TYPE(SCHEME_TYPE(x), scheme_compiled_unquote_splicing_type)) {
    list = _scheme_eval_compiled(((Scheme_Object *)SCHEME_PTR1_VAL(x)), env);
    rest = quasi_execute((Scheme_Object *)SCHEME_PTR2_VAL(x), env);
    if (SCHEME_NULLP(rest))
      return list;
    else
      return scheme_append(list, rest);
  }

  if (!SCHEME_PAIRP(x))
    return x;

  qcar = quasi_execute(SCHEME_CAR (x), env);
  qcdr = quasi_execute(SCHEME_CDR (x), env);
  return scheme_make_pair (qcar, qcdr);
}

static Scheme_Object *
quasi_syntax (int edepth,
	      Scheme_Object *x, int level, Scheme_Env *env,
	      Scheme_Compile_Info *rec)
{
  Scheme_Object *form, *rest, *cell, *qcar, *qcdr, *ret;
  int cdr_level, car_level;
  Scheme_Compile_Info recs[2];

  if (rec) {
    rec->max_let_depth = 0;
    rec->can_return_closure = 0;
  }

  if (SCHEME_VECTORP (x)) {
    x = scheme_vector_to_list (x);
    x = quasi_syntax(edepth, x, level, env, rec);

    if (!rec)
      return scheme_list_to_vector(x);
    else {
      ret = scheme_alloc_stubborn_small_object();
      ret->type = scheme_compiled_quoted_vector_type;
      SCHEME_PTR_VAL(ret) = x;
      scheme_end_stubborn_change((void *)ret);
    
      return ret;
    }
  }

  if (!SCHEME_PAIRP(x))
    return x;

  cdr_level = car_level = level;

  if (SAME_OBJ(SCHEME_CAR(x), scheme_unquote)) {
    if (level)
      --cdr_level;
    else {
      x = SCHEME_CDR (x);
      SCHEME_ASSERT(SCHEME_PAIRP(x) && SCHEME_NULLP(SCHEME_CDR(x)), 
		    "bad unquote form");
      
      if (!rec)
	return cons(scheme_unquote, scheme_expand_expr(x, env, edepth));
      else {
	cell = scheme_compile_expr(SCHEME_CAR (x), env, rec);
	
	ret = scheme_alloc_stubborn_small_object();
	ret->type = scheme_compiled_unquote_type;
	SCHEME_PTR_VAL(ret) = cell;
	scheme_end_stubborn_change((void *)ret);
	
	return ret;
      }
    }
  } else if (SCHEME_PAIRP(SCHEME_CAR (x))
	   && SAME_OBJ(SCHEME_CAR(SCHEME_CAR (x)), scheme_unquote_splicing)) {
    if (level)
      --car_level;
    else {
      Scheme_Object *a, *b;

      SCHEME_ASSERT (SCHEME_NULLP(SCHEME_CDDR(SCHEME_CAR(x))), 
		     "bad unquote-splicing form");

      a = SCHEME_CADR(SCHEME_CAR(x));
      b = SCHEME_CDR(x);

      if (rec)
	scheme_init_compile_recs(rec, recs, 2);
	
      rest = quasi_syntax(edepth, b, cdr_level, env, 
			  rec ? (Scheme_Compile_Info *)&recs[1] : rec);
	
      if (!rec) {
	return cons(cons(scheme_unquote_splicing,
			 cons(scheme_expand_expr(a, env, edepth),
			      scheme_null)),
		    rest);
      } else {
	form = scheme_compile_expr(a, env, recs);
	
	scheme_merge_compile_recs(rec, recs, 2);
	
	ret = scheme_alloc_stubborn_object();
	ret->type = scheme_compiled_unquote_splicing_type;
	SCHEME_PTR1_VAL(ret) = form;
	SCHEME_PTR2_VAL(ret) = rest;
	scheme_end_stubborn_change((void *)ret);
	
	return ret;
      }
    } 
  } else if (SAME_OBJ(SCHEME_CAR(x),  scheme_quasiquote))
    ++cdr_level;

  if (!rec)
    return cons(quasi_syntax(edepth, SCHEME_CAR(x), car_level, env, rec),
		quasi_syntax(edepth, SCHEME_CDR(x), car_level, env, rec));
  else {
    scheme_init_compile_recs(rec, recs, 2);
    
    qcar = quasi_syntax(edepth, SCHEME_CAR (x), car_level, env, recs);
    qcdr = quasi_syntax(edepth, SCHEME_CDR (x), cdr_level, env, recs + 1);
    
    scheme_merge_compile_recs(rec, recs, 2);
    
    return scheme_make_pair (qcar, qcdr);
  }
}

#endif

static Scheme_Object *
defmacro_execute (Scheme_Object *form, Scheme_Env *env)
{
  Scheme_Object *name;
  Scheme_Object *fun, *macro;

  name = SCHEME_CAR(form);
  fun = SCHEME_CDR(form);

  fun = _scheme_eval_compiled(fun, env);

  macro = scheme_alloc_stubborn_small_object ();
  macro->type = scheme_macro_type;
  SCHEME_PTR_VAL (macro) = fun;

  scheme_end_stubborn_change((void *)macro);

  scheme_add_global (SCHEME_SYM_VAL(name), macro, env);

  return scheme_void;
}

static void defmacro_parse(Scheme_Object *form, 
			   Scheme_Object **name, Scheme_Object **code)
{
  SCHEME_ASSERT((scheme_list_length (form) > 3), "badly formed defmacro");
  *name = SCHEME_CAR (SCHEME_CDR (form));
  if (!SCHEME_SYMBOLP(*name))
    scheme_wrong_type("defmacro", "symbol", -1, 0, name);
  *code = SCHEME_CDR (SCHEME_CDR (form));
}

static Scheme_Object *
defmacro_syntax (Scheme_Object *form, Scheme_Env *env, 
		 Scheme_Compile_Info *rec)
{
  Scheme_Object *name, *code;
  Scheme_Object *fun;

  if (env->basic.next)
    scheme_signal_error("defmacro: only allowed at top-level");

  defmacro_parse(form, &name, &code);

  fun = scheme_make_closure_compilation(env, code, rec);

  rec->max_let_depth = 0;
  rec->can_return_closure = 1;

  return scheme_make_syntax_compilation(defmacro_execute,
					scheme_make_pair(name, fun));
}

static Scheme_Object *
defmacro_expand(Scheme_Object *form, Scheme_Env *env, int depth)
{
  Scheme_Object *name, *code, *fpart;

  defmacro_parse(form, &name, &code);

  fpart = SCHEME_CDR(lambda_expand(cons(scheme_lambda, code), env, depth));

  return cons(SCHEME_CAR(form), cons(name, fpart));
}

static Scheme_Object *
do_letmacro(Scheme_Object *forms, Scheme_Env *env, 
	    Scheme_Compile_Info *rec, int depth)
{
  Scheme_Object *form, *name, *args, *mbody, *body, *cl, *v;
  Scheme_Object *oldv, *macro;
  Scheme_Bucket *b;
  jmp_buf save;

  /* Implemented sortof like using fluid-let */

  form = SCHEME_CDR(forms);
  SCHEME_ASSERT(SCHEME_PAIRP(form), "badly formed 'letmacro'");
  name = SCHEME_CAR(form);
  form = SCHEME_CDR(form);
  SCHEME_ASSERT(SCHEME_PAIRP(form), "badly formed 'letmacro'");
  args = SCHEME_CAR(form);
  form = SCHEME_CDR(form);
  SCHEME_ASSERT(SCHEME_PAIRP(form), "badly formed 'letmacro'");
  mbody = SCHEME_CAR(form);
  body = SCHEME_CDR(form);
  
  if (!SCHEME_SYMBOLP(name))
    scheme_wrong_type("letmacro", "symbol", -1, 0, &name);

  cl = cons(scheme_lambda,
	    cons(args,
		 cons(mbody, scheme_null)));

  cl = scheme_eval(cl, scheme_env);

  macro = scheme_alloc_stubborn_small_object ();
  macro->type = scheme_macro_type;
  SCHEME_PTR_VAL (macro) = cl;
  scheme_end_stubborn_change((void *)macro);

  b = scheme_global_bucket(name, env);
  oldv = (Scheme_Object *)b->val;
  b->val = (void *)macro;
  memcpy(&save, &scheme_error_buf, sizeof(jmp_buf));
  if (setjmp(scheme_error_buf)) {
    b->val = (void *)oldv;
    longjmp(save, 1);
  } else {
    if (rec) {
      v = scheme_compile_block(body, env, rec);
      v = scheme_make_sequence_compilation(v);
    } else
      v = cons(SCHEME_CAR(forms),
	       cons(name,
		    cons(args,
			 cons(mbody, scheme_expand_block(body, env, depth)))));
  }

  memcpy(&scheme_error_buf, &save, sizeof(jmp_buf));
  b->val = (void *)oldv;

  return v;
}

static Scheme_Object *
letmacro_syntax(Scheme_Object *form, Scheme_Env *env, 
		Scheme_Compile_Info *rec)
{
  return do_letmacro(form, env, rec, 0);
}

static Scheme_Object *
letmacro_expand(Scheme_Object *form, Scheme_Env *env, int depth)
{
  return do_letmacro(form, env, NULL, depth);
}

static Scheme_Object *void_execute(Scheme_Object *expr, Scheme_Env *env)
{
  return scheme_void;
}

/*********************************************************************/

typedef struct {
  Scheme_Object *sym;
  Scheme_Syntax_Compilation *f;
} Executer_Name;

static int num_exec_names = 0;
static Executer_Name *executer_names;

void scheme_register_syntax_executer(const char *name, 
				     Scheme_Syntax_Compilation *f)
{
  Executer_Name *old = executer_names;

  executer_names = 
    (Executer_Name *)scheme_malloc_atomic((num_exec_names + 1)
					  * sizeof(Executer_Name));

  memcpy(executer_names, old, num_exec_names * sizeof(Executer_Name));

  executer_names[num_exec_names].sym = scheme_intern_symbol(name);
  executer_names[num_exec_names].f = f;
  num_exec_names++;
}

Scheme_Object *scheme_find_executer_name(Scheme_Syntax_Compilation *f)
{
  int i;

  for (i = 0; i < num_exec_names; i++)
    if (SAME_PTR(executer_names[i].f, f))
      return executer_names[i].sym;

  return NULL;
}

Scheme_Syntax_Compilation *scheme_find_executer(Scheme_Object *sym)
{
  int i;

  for (i = 0; i < num_exec_names; i++)
    if (SAME_OBJ(executer_names[i].sym, sym))
      return executer_names[i].f;

  return NULL;
}

/*****************************************************************/

#define BOOL(x) (x ? scheme_true : scheme_false)

static Scheme_Object *write_let(Scheme_Object *obj)
{
  Let_Data *data;
  
  data = (Let_Data *)obj;

  return cons(scheme_make_integer(data->num_bindings), data->body);
}

static Scheme_Object *read_let(Scheme_Object *obj)
{
  Let_Data *data;

  data = (Let_Data *)scheme_malloc_stubborn(sizeof(Let_Data));

  data->type = scheme_let_data_type;

  data->num_bindings = SCHEME_INT_VAL(SCHEME_CAR(obj));
  data->body = SCHEME_CDR(obj);

  scheme_end_stubborn_change((void *)data);

  return (Scheme_Object *)data;
}

static Scheme_Object *write_let_value(Scheme_Object *obj)
{
  Scheme_Let_Value *lv;
 
  lv = (Scheme_Let_Value *)obj;

  return cons(scheme_make_integer(lv->position),
	      cons(lv->value, lv->body));
}

static Scheme_Object *read_let_value(Scheme_Object *obj)
{
  Scheme_Let_Value *lv;
 
  lv = (Scheme_Let_Value *)scheme_malloc(sizeof(Scheme_Let_Value));
  lv->type = scheme_let_value_compilation_type;

  lv->position = SCHEME_INT_VAL(SCHEME_CAR(obj));
  obj = SCHEME_CDR(obj);
  lv->value = SCHEME_CAR(obj);
  lv->body = SCHEME_CDR(obj);

  return (Scheme_Object *)lv;
}

static Scheme_Object *write_let_void(Scheme_Object *obj)
{
  Scheme_Let_Void *lv;
 
  lv = (Scheme_Let_Void *)obj;

  return cons(scheme_make_integer(lv->start), 
	      cons(scheme_make_integer(lv->end), lv->body));
}

static Scheme_Object *read_let_void(Scheme_Object *obj)
{
  Scheme_Let_Void *lv;
 
  lv = (Scheme_Let_Void *)scheme_malloc(sizeof(Scheme_Let_Void));
  lv->type = scheme_let_void_compilation_type;

  lv->start = SCHEME_INT_VAL(SCHEME_CAR(obj));
  obj = SCHEME_CDR(obj);
  lv->end = SCHEME_INT_VAL(SCHEME_CAR(obj));
  lv->body = SCHEME_CDR(obj);

  return (Scheme_Object *)lv;
}

static Scheme_Object *write_uq(Scheme_Object *obj)
{
  return cons((Scheme_Object *)SCHEME_PTR_VAL(obj), scheme_null);
}

static Scheme_Object *read_uq(Scheme_Object *obj)
{
  Scheme_Object *ret;

  if (!SCHEME_PAIRP(obj) || !SCHEME_NULLP(SCHEME_CDR(obj)))
    scheme_signal_error("bad compiled unquote");
  
  ret = scheme_alloc_stubborn_small_object();
  ret->type = scheme_compiled_unquote_type;
  SCHEME_PTR_VAL(ret) = SCHEME_CAR(obj);
  scheme_end_stubborn_change((void *)ret);

  return ret;
}

static Scheme_Object *write_uq_splice(Scheme_Object *obj)
{
  return cons((Scheme_Object *)SCHEME_PTR1_VAL(obj),
	      (Scheme_Object *)SCHEME_PTR2_VAL(obj));
}

static Scheme_Object *read_uq_splice(Scheme_Object *obj)
{
  Scheme_Object *ret;

  if (!SCHEME_PAIRP(obj))
    scheme_signal_error("bad compiled unquote-splicing");
  
  ret = scheme_alloc_stubborn_object();
  ret->type = scheme_compiled_unquote_splicing_type;
  SCHEME_PTR1_VAL(ret) = SCHEME_CAR(obj);
  SCHEME_PTR2_VAL(ret) = SCHEME_CDR(obj);
  scheme_end_stubborn_change((void *)ret);

  return ret;
}

static Scheme_Object *write_quoted_vector(Scheme_Object *obj)
{
  return cons((Scheme_Object *)SCHEME_PTR_VAL(obj), scheme_null);
}

static Scheme_Object *read_quoted_vector(Scheme_Object *obj)
{
  Scheme_Object *ret;

  if (!SCHEME_PAIRP(obj) || !SCHEME_NULLP(SCHEME_CDR(obj)))
    scheme_signal_error("bad compiled quoted vector");
  
  ret = scheme_alloc_stubborn_small_object();
  ret->type = scheme_compiled_quoted_vector_type;
  SCHEME_PTR_VAL(ret) = SCHEME_CAR(obj);
  scheme_end_stubborn_change((void *)ret);

  return ret;
}

