/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"

/* global_constants */
/* locals */
static Scheme_Object *vector_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *make_vector (int argc, Scheme_Object *argv[]);
static Scheme_Object *vector (int argc, Scheme_Object *argv[]);
static Scheme_Object *vector_length (int argc, Scheme_Object *argv[]);
static Scheme_Object *vector_ref (int argc, Scheme_Object *argv[]);
static Scheme_Object *vector_set (int argc, Scheme_Object *argv[]);
static Scheme_Object *vector_to_list (int argc, Scheme_Object *argv[]);
static Scheme_Object *list_to_vector (int argc, Scheme_Object *argv[]);
static Scheme_Object *vector_fill (int argc, Scheme_Object *argv[]);

static Scheme_Object *vector_append (int argc, Scheme_Object *argv[]);

void
scheme_init_vector (Scheme_Env *env)
{
  scheme_add_global_constant ("vector?", scheme_make_prim_w_arity (vector_p, 1, 1), env);
  scheme_add_global_constant ("make-vector", scheme_make_prim_w_arity (make_vector, 1, 2), env);
  scheme_add_global_constant ("vector", scheme_make_prim_w_arity (vector, 0, -1), env);
  scheme_add_global_constant ("vector-length", scheme_make_prim_w_arity (vector_length, 1, 1), env);
  scheme_add_global_constant ("vector-ref", scheme_make_prim_w_arity (vector_ref, 2, 2), env);
  scheme_add_global_constant ("vector-set!", scheme_make_prim_w_arity (vector_set, 3, 3), env);
  scheme_add_global_constant ("vector->list", scheme_make_prim_w_arity (vector_to_list, 1, 1), env);
  scheme_add_global_constant ("list->vector", scheme_make_prim_w_arity (list_to_vector, 1, 1), env);
  scheme_add_global_constant ("vector-fill!", scheme_make_prim_w_arity (vector_fill, 2, 2), env);

  scheme_add_global_constant ("vector-append", scheme_make_prim_w_arity (vector_append, 1, -1), env);
}

Scheme_Object *
scheme_make_vector (int size, Scheme_Object *fill)
{
  Scheme_Object *vec;
  int i;

  if (size < 0)
    scheme_signal_error("make-vector: cannot allocate negative size");

  vec = scheme_alloc_stubborn_object ();
  vec->type = scheme_vector_type;
  SCHEME_VEC_SIZE(vec) = size;
  SCHEME_VEC_ELS(vec) = (Scheme_Object**)scheme_malloc(sizeof(Scheme_Object*)*size);
  for ( i=0 ; i<size ; ++i )
    {
      SCHEME_VEC_ELS(vec)[i] = fill;
    }

  scheme_end_stubborn_change((void *)vec);

  return (vec);
}

/* locals */

static Scheme_Object *
vector_p (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("vector?", 1, 1, argc, argv);

  return (SCHEME_VECTORP(argv[0]) ? scheme_true : scheme_false);
}

static Scheme_Object *
make_vector (int argc, Scheme_Object *argv[])
{
  Scheme_Object *vec, *fill;
  int len;

  if (argc != 1 && argc != 2)
    scheme_wrong_count("make-vector", 1, 2, argc, argv);
  if (!SCHEME_INTP(argv[0]) && !SCHEME_DBLP(argv[0]))
    scheme_wrong_type("make-vector", "number", 0, argc, argv);

  len = SCHEME_NUM_VAL (argv[0]);
  if (argc == 2)
    fill = argv[1];
  else
    fill = scheme_false;

  vec = scheme_make_vector (len, fill);

  return vec;
}

static Scheme_Object *
vector (int argc, Scheme_Object *argv[])
{
  Scheme_Object *vec;
  int i;

  vec = scheme_make_vector (argc, 0);
  for ( i=0 ; i<argc ; ++i )
    SCHEME_VEC_ELS(vec)[i] = argv[i];

  return (vec);
}

static Scheme_Object *
vector_length (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("vector-length", 1, 1, argc, argv);
  if (!SCHEME_VECTORP(argv[0]))
    scheme_wrong_type("vector-length", "vector", 0, argc, argv);

  return (scheme_make_integer (SCHEME_VEC_SIZE (argv[0])));
}

static Scheme_Object *
vector_ref (int argc, Scheme_Object *argv[])
{
  int i;

  if (argc != 2)
    scheme_wrong_count("vector-ref", 2, 2, argc, argv);
  if (!SCHEME_VECTORP(argv[0]))
    scheme_wrong_type("vector-ref", "vector", 0, argc, argv);

  if (SCHEME_INTP (argv[1]))
    i = SCHEME_INT_VAL (argv[1]);
  else if (SCHEME_DBLP (argv[1]))
    i = SCHEME_DBL_VAL (argv[1]);
  else {
    scheme_wrong_type("vector-ref", "number", 1, argc, argv);
    i = 0;
  }
  
  if (!((i >= 0) && (i < SCHEME_VEC_SIZE (argv[0]))))
    scheme_signal_error("vector-ref: index %d out of range [%d, %d]",
			i, 0, SCHEME_VEC_SIZE (argv[0]) - 1);

  return (SCHEME_VEC_ELS(argv[0])[i]);
}

static Scheme_Object *
vector_set (int argc, Scheme_Object *argv[])
{
  int i;

  if (argc != 3)
    scheme_wrong_count("vector-set!", 3, 3, argc, argv);
  if (!SCHEME_VECTORP(argv[0]))
    scheme_wrong_type("vector-set!", "vector", 0, argc, argv);

  if (SCHEME_INTP (argv[1]))
    i = SCHEME_INT_VAL (argv[1]);
  else if (SCHEME_DBLP (argv[1]))
    i = SCHEME_DBL_VAL (argv[1]);
  else {
    scheme_wrong_type("vector-set!", "number", 1, argc, argv);
    i = 0;
  }

  if (!((i >= 0) && (i < SCHEME_VEC_SIZE (argv[0]))))
    scheme_signal_error("vector-set!: index %d out of range [%d, %d]",
			i, 0, SCHEME_VEC_SIZE (argv[0]) - 1);

  SCHEME_VEC_ELS(argv[0])[i] = argv[2];

  return scheme_void;
}

static Scheme_Object *
vector_to_list (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("vector->list", 1, 1, argc, argv);
  if (!SCHEME_VECTORP(argv[0]))
    scheme_wrong_type("vector->list", "vector", 0, argc, argv);

  return scheme_vector_to_list(argv[0]);
}

Scheme_Object *
scheme_vector_to_list (Scheme_Object *vec)
{
  int len, i;
  Scheme_Object *first, *last, *pair;

  len = SCHEME_VEC_SIZE (vec);
  first = last = scheme_null;
  for ( i=0 ; i<len ; ++i )
    {
      pair = scheme_make_pair (SCHEME_VEC_ELS(vec)[i], scheme_null);
      if (SCHEME_NULLP(first))
	{
	  first = last = pair;
	}
      else
	{
	  SCHEME_CDR (last) = pair;
	  last = pair;
	}
    }
  return (first);
}

static Scheme_Object *
list_to_vector (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("list->vector", 1, 1, argc, argv);

  return scheme_list_to_vector(argv[0]);
}

Scheme_Object *
scheme_list_to_vector (Scheme_Object *list)
{
  int len, i;
  Scheme_Object *vec, *orig = list;

  len = scheme_list_length (list);
  vec = scheme_make_vector (len, 0);
  i = 0;
  while (SCHEME_PAIRP(list)) {
    SCHEME_VEC_ELS(vec)[i] = SCHEME_CAR (list);
    i++;
    list = SCHEME_CDR (list);
  }

  if (!SCHEME_NULLP(list))
      scheme_wrong_type("list->vector", "proper list", -1, 0, &orig);

  return vec;
}

static Scheme_Object *
vector_fill (int argc, Scheme_Object *argv[])
{
  int i;
  
  if (argc != 2)
    scheme_wrong_count("vector-fill!", 2, 2, argc, argv);
  if (!SCHEME_VECTORP(argv[0]))
    scheme_wrong_type("vector-fill!", "vector", 0, argc, argv);

  for ( i=0 ; i<SCHEME_VEC_SIZE (argv[0]) ; ++i )
    SCHEME_VEC_ELS(argv[0])[i] = argv[1];

  return argv[0];
}

static Scheme_Object *
vector_append (int argc, Scheme_Object *argv[])
{
  int len1, len2, i;
  Scheme_Object *naya;

  if (argc != 2)
    scheme_wrong_count("vector-append", 2, 2, argc, argv);
  if (!SCHEME_VECTORP(argv[0]))
    scheme_wrong_type("vector-append", "vector", 0, argc, argv);
  if (!SCHEME_VECTORP(argv[1]))
    scheme_wrong_type("vector-append", "vector", 1, argc, argv);


  len1 = SCHEME_VEC_SIZE (argv[0]);
  len2 = SCHEME_VEC_SIZE (argv[1]);
  naya = scheme_make_vector ((len1 + len2), NULL);

  for ( i=0 ; i<len1 ; ++i )
    SCHEME_VEC_ELS(naya)[i] = SCHEME_VEC_ELS(argv[0])[i];

  for ( i=len1 ; i<(len1 + len2) ; ++i )
    SCHEME_VEC_ELS(naya)[i] = SCHEME_VEC_ELS(argv[1])[i-len1];

  return naya;
}
